# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Dict, Iterable, List, Set
from . import CodeRange


__all__ = [
  'codes_desc',
]


def codes_desc(code_ranges:Iterable[CodeRange], raw=False) -> str:
  if raw:
    return ' '.join(codes_range_desc(*p) for p in code_ranges) or 'Ø'
  # Calculate single-character subtraction simplifications for ranges.
  ranges:List[CodeRange] = []
  subtracted:Set[int] = set()
  for r in code_ranges: # Assumed to be sorted.
    if ranges:
      p = ranges[-1]
      if p[1] + 1 == r[0] and p[1] != 0x60: # Exclude '`' because it makes standard sym range look weird.
        ranges[-1] = (p[0], r[1])
        subtracted.add(p[1])
      else:
        ranges.append(r)
    else:
      ranges.append(r)
  if not ranges: return 'Ø'
  rs = ' '.join(codes_range_desc(*p) for p in ranges)
  if not subtracted: return rs
  ss = ' '.join(code_desc(c) for c in subtracted)
  return f'{rs} - {ss}'


def codes_range_desc(l:int, h:int) -> str:
  if l + 1 == h: return code_desc(l)
  return '{}..{}'.format(code_desc(l), code_desc(h-1))

def code_desc(c:int) -> str:
  assert isinstance(c, int)
  try: return code_descriptions[c]
  except KeyError: return f'\\{c:02x}/'

code_descriptions:Dict[int, str] = {
  -1: 'Ø',
  ord('\a'): '\\a',
  ord('\b'): '\\b',
  ord('\t'): '\\t',
  ord('\n'): '\\n',
  ord('\v'): '\\v',
  ord('\f'): '\\f',
  ord('\r'): '\\r',
  ord(' '): '\\s',
}

code_descriptions.update((i, chr(i)) for i in range(ord('!'), 0x7f))



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
HTML semantics data.
'''

raw_text_tags = frozenset({ 'script', 'style' })
escapeable_raw_text_tags = frozenset({ 'textarea', 'title' })
ws_sensitive_tags = raw_text_tags | escapeable_raw_text_tags | frozenset({'pre'})

void_tags = frozenset({
  'area',
  'base',
  'br',
  'col',
  'embed',
  'hr',
  'img',
  'input',
  'link',
  'meta',
  'param',
  'source',
  'track',
  'wbr',
})

form_input_types = frozenset({
  'button', # Push button with no default behavior.
  'checkbox', # Check box allowing single values to be selected/deselected.
  'color', # Control for specifying a color. A color picker's UI has no required features other than accepting simple colors as text (more info).
  'date', # Control for entering a date (year, month, and day, with no time).
  'datetime-local', # Control for entering a date and time, with no time zone.
  'email', # Field for editing an e-mail address.
  'file', # Control that lets the user select a file. Use `accept` to define the types of files that the control can select.
  'hidden', # Control that is not displayed but whose value is submitted to the server.
  'image', # Graphical submit button. `src` specifies the image and `alt` specifies alternative text. Use the height and width attributes to define the size of the image in pixels.
  'month', # Control for entering a month and year, with no time zone.
  'number', # Control for entering a number.
  'password', # Single-line text field whose value is obscured. Use the maxlength and minlength attributes to specify the maximum length of the value that can be entered.
  'radio', # Radio button, allowing a single value to be selected out of multiple choices.
  'range', # Control for entering a number whose exact value is not important.
  'reset', # Button that resets the contents of the form to default values.
  'search', # Single-line text field for entering search strings. Line-breaks are automatically removed from the input value.
  'submit', # Button that submits the form.
  'tel', # Control for entering a telephone number.
  'text', # Single-line text field. Line-breaks are automatically removed from the input value.
  'time', # Control for entering a time value with no time zone.
  'url', # Field for entering a URL.
  'week', # Control for entering a date consisting of a week-year number and a week number with no time zone.
})


phrasing_tags = frozenset({
  'a',
  'abbr',
  'area',
  'audio',
  'b',
  'bdi',
  'bdo',
  'br',
  'button',
  'canvas',
  'cite',
  'code',
  'data',
  'datalist',
  'del',
  'dfn',
  'em',
  'embed',
  'i',
  'iframe',
  'img',
  'input',
  'ins',
  'kbd',
  'label',
  'map',
  'mark',
  'meta',
  'meter',
  'noscript',
  'object',
  'output',
  'picture',
  'progress',
  'q',
  'ruby',
  's',
  'samp',
  'script',
  'select',
  'slot',
  'small',
  'span',
  'strong',
  'sub',
  'sup',
  'template',
  'textarea',
  'time',
  'u',
  'var',
  'video',
  'wbr',
})



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.


from enum import Enum
from urllib.parse import urlsplit, urlunsplit

from .path import norm_path, path_join, path_split


def path_encode(string:str) -> str:
  '''
  Encode string into a path.
  Path encoding is similar to url percent encoding, but uses '+' as the indicator character.
  It is used to create convenient file names out of URLs.
  Like URL encoding, letters, digits, and the characters '_.-' are left unescaped.
  Additionally:
  * A leading '.' is escaped to prevent hidden files from being created.
  * '%' characters are not escaped, to make prior URL encoding more readable.
  * ':' characters are not escaped to make URL schemes more readable.
  * '=' characters are not escaped.
  * '/' is translated to ',', which allows multiple path segments in the url to be squashed into a single filename.
  * Note that ',' is itself encoded, so the encoding is unambiguous.
  * All other bytes are encoded as "+XX", where XX is the uppercase hexadecimal byte value.
  '''
  utf8 = string.encode()
  path = ''.join(_path_encode_table[b] for b in utf8)
  if path.startswith('.'):
    return _encode_dot + path[1:]
  else:
    return path


_encode_dot = f'+{ord("."):02x}'

_path_encode_table = [f'+{i:02x}' for i in range(0x100)]

_path_encode_table[ord('/')] = ','

for c in '%-._:=':
  _path_encode_table[ord(c)] = c

for start, last in ['09', 'AZ', 'az']:
  for code in range(ord(start), ord(last) + 1):
    _path_encode_table[code] = chr(code)


class PartMode(Enum):
  OMIT = 0
  SQUASH = 1
  COMP = 2
  SPLIT = 3

OMIT = PartMode.OMIT
SQUASH = PartMode.SQUASH
COMP = PartMode.COMP
SPLIT = PartMode.SPLIT

def path_for_url(url:str, *, normalize=True,
 scheme=OMIT, host=COMP, path=COMP, query=SQUASH, fragment=OMIT) -> str:
  '''
  Return a path encoded from a url by first splitting the url into parts
  and then applying `path_encode` to each part.
  Empty parts are represented by a bare '+'.

  `split_path`: if True, the path is translated into nested directories.
  Otherwise, slashes are translated into commas.

  `scheme`, `host`, `path`, `query`, `fragment` are all PartMode options.
  * `OMIT` causes that part to be omitted;
  * `SQUASH` causes it to be squashed into the previous path component;
  * `COMP` causes it to be appended as a distinct path component;
  * `SPLIT` is only valid for `path`, and splits the path into components.
  '''

  if scheme in (SQUASH, SPLIT): raise ValueError(f'`scheme` mode cannot be `{scheme.name}`')
  if host == SPLIT: raise ValueError('`host` mode cannot be `SPLIT`')
  if query == SPLIT: raise ValueError('`query` mode cannot be `SPLIT`')
  if fragment == SPLIT: raise ValueError('`fragment` mode cannot be `SPLIT`')


  parts = urlsplit(url) # returns five-element namedtuple.
  comps = []
  def add_comp(mode:PartMode, part:str) -> None:
    if mode is OMIT:
      return
    elif mode is COMP:
      comps.append(part)
    elif mode is SQUASH:
      if comps:
        comps[-1] += part
      else:
        comps.append(part)
    elif mode is SPLIT:
      comps.extend(path_split(part) if part else ['']) # Do not normalize '' -> '.'.
    else:
      raise NotImplementedError

  add_comp(scheme, parts.scheme + ':') # Always include the trailing colon for scheme.

  # The only case where we can omit the host '//' is if neither host nor path is squashed;
  # then host always has its own directory level and the slashes are just noise.
  h = parts.netloc if (host != SQUASH and path != SQUASH and parts.netloc) else ('//' + parts.netloc)
  add_comp(host, h)

  p = parts.path
  if p and normalize: p = norm_path(parts.path) # Do not normalize '' -> '.'.
  # If host has its own directory level and is nonempty then path's leading slash can be omitted, because it must be present.
  if host == COMP and path != SQUASH and parts.netloc and p and p[0] == '/': p = p[1:]
  add_comp(path, p)

  if parts.query:
    add_comp(query, '?' + parts.query)
  if parts.fragment:
    add_comp(fragment, '#' + parts.fragment)

  res = path_join(*[path_encode(c) if c else '+' for c in comps])
  assert res and res[0] != '/', res
  return res


def norm_url(url:str) -> str:
  scheme, host, path, query, fragment = urlsplit(url)
  return urlunsplit((scheme, host, norm_path(path), query, fragment))



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

# rename files to have lower case extensions.

from argparse import ArgumentParser
from sys import stderr

from ..fs import file_inode, move_file, walk_files
from ..path import split_stem_ext


def main() -> None:
  parser = ArgumentParser(description='Lower case file extensions.')
  parser.add_argument('paths', nargs='+', help='Paths to explore.')
  args = parser.parse_args()

  for path in walk_files(*args.paths):
    stem, ext = split_stem_ext(path)
    lc_ext = ext.lower()
    if lc_ext == ext:
      continue
    lc_path = stem + lc_ext
    try: inode = file_inode(lc_path, follow=False)
    except FileNotFoundError: pass
    else:
      # OSX file system is case insensitive, so lc_path always appears to exist.
      # Check if it is really the same file by comparing inodes.
      if file_inode(path, follow=False) != inode:
        print('error: {!r} -> {!r}; destination already exists.'.format(path, lc_path), file=stderr)
        continue
    move_file(path, to=lc_path, overwrite=True)


if __name__ == '__main__': main()




module_names = [
  '_dummy_thread',
  '_thread',
  'abc',
  'aifc',
  'argparse',
  'array',
  'ast',
  'asynchat',
  'asyncio',
  'asyncore',
  'atexit',
  'audioop',
  'base64',
  'bdb',
  'binascii',
  'binhex',
  'bisect',
  'builtins',
  'bz2',
  'calendar',
  'cgi',
  'cgitb',
  'chunk',
  'cmath',
  'cmd',
  'code',
  'codecs',
  'codeop',
  'collections.abc',
  'collections',
  'colorsys',
  'compileall',
  'concurrent.futures',
  'configparser',
  'contextlib',
  'contextvars',
  'copy',
  'copyreg',
  'crypt',
  'csv',
  'ctypes',
  'curses.ascii',
  'curses.panel',
  'curses.textpad',
  'curses',
  'dataclasses',
  'datetime',
  'dbm',
  'decimal',
  'difflib',
  'dis',
  'distutils',
  'doctest',
  'dummy_threading',
  'email',
  'ensurepip',
  'enum',
  'errno',
  'faulthandler',
  'fcntl',
  'filecmp',
  'fileinput',
  'fnmatch',
  'formatter',
  'fractions',
  'ftplib',
  'functools',
  'gc',
  'getopt',
  'getpass',
  'gettext',
  'glob',
  'grp',
  'gzip',
  'hashlib',
  'heapq',
  'hmac',
  'html.entities',
  'html.parser',
  'html',
  'http.client',
  'http.cookiejar',
  'http.cookies',
  'http.server',
  'http',
  'imaplib',
  'imghdr',
  'imp',
  'importlib',
  'inspect',
  'io',
  'ipaddress',
  'itertools',
  'json',
  'keyword',
  'linecache',
  'locale',
  'logging.config',
  'logging.handlers',
  'logging',
  'lzma',
  'macpath',
  'mailbox',
  'mailcap',
  'marshal',
  'math',
  'mimetypes',
  'mmap',
  'modulefinder',
  'msilib',
  'msvcrt',
  'multiprocessing',
  'netrc',
  'nis',
  'nntplib',
  'numbers',
  'operator',
  'optparse',
  'os.path',
  'os',
  'ossaudiodev',
  'parser',
  'pathlib',
  'pdb',
  'pickle',
  'pickletools',
  'pipes',
  'pkgutil',
  'platform',
  'plistlib',
  'poplib',
  'posix',
  'pprint',
  'pty',
  'pwd',
  'py_compile',
  'pyclbr',
  'pydoc',
  'queue',
  'quopri',
  'random',
  're',
  'readline',
  'reprlib',
  'resource',
  'rlcompleter',
  'runpy',
  'sched',
  'secrets',
  'select',
  'selectors',
  'shelve',
  'shlex',
  'shutil',
  'signal',
  'site',
  'site',
  'smtpd',
  'smtplib',
  'sndhdr',
  'socket',
  'socketserver',
  'spwd',
  'sqlite3',
  'ssl',
  'stat',
  'statistics',
  'string',
  'stringprep',
  'struct',
  'subprocess',
  'sunau',
  'symbol',
  'symtable',
  'sys',
  'sysconfig',
  'syslog',
  'tabnanny',
  'tarfile',
  'telnetlib',
  'tempfile',
  'termios',
  'test.support.script_helper',
  'test.support',
  'test',
  'textwrap',
  'threading',
  'time',
  'timeit',
  'tkinter.scrolledtext',
  'tkinter.tix',
  'tkinter.ttk',
  'tkinter',
  'token',
  'tokenize',
  'trace',
  'traceback',
  'tracemalloc',
  'tty',
  'turtle',
  'types',
  'typing',
  'unicodedata',
  'unittest.mock',
  'unittest.mock',
  'unittest',
  'urllib.error',
  'urllib.parse',
  'urllib.request',
  'urllib.response',
  'urllib.robotparser',
  'urllib',
  'uu',
  'uuid',
  'venv',
  'warnings',
  'wave',
  'weakref',
  'webbrowser',
  'winreg',
  'winsound',
  'wsgiref',
  'xdrlib',
  'xml.dom.minidom',
  'xml.dom.pulldom',
  'xml.dom',
  'xml.etree.ElementTree',
  'xml.parsers.expat',
  'xml.sax.handler',
  'xml.sax.saxutils',
  'xml.sax.xmlreader',
  'xml.sax',
  'xmlrpc.client',
  'xmlrpc.server',
  'xmlrpc',
  'zipapp',
  'zipfile',
  'zipimport',
  'zlib',
]



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from argparse import ArgumentParser
from html.parser import HTMLParser
from sys import stdin
from typing import List, Optional, Tuple


def main() -> None:
  arg_parser = ArgumentParser('Check the validity of HTML documents.')
  arg_parser.add_argument('paths', nargs='*', help='paths to HTML documents (defaults to stdin).')
  args = arg_parser.parse_args()

  try: files = [open(path) for path in args.paths]
  except FileNotFoundError as e: exit(f'file not found: {e.filename}')
  if not files: files = [stdin]

  for file in files:
    parser = Parser(path=file.name)
    parser.feed(file.read())
    parser.close()
    parser.check_completeness()


Pos = Tuple[int, int]


class Parser(HTMLParser):

  def __init__(self, path: str) -> None:
    super().__init__(convert_charrefs=True)
    self.path = path
    self.stack: List[Tuple[Pos, str]] = []
    self.found_leading_doctype = False

  def check_completeness(self) -> None:
    if not self.found_leading_doctype:
      self.msg("did not find '<!DOCTYPE html>' declaration in leading position.", pos=(0,0))

  def handle_startendtag(self, tag: str, attrs: List[Tuple[str, Optional[str]]]) -> None:
    pass

  def handle_starttag(self, tag: str, attrs: List[Tuple[str,Optional[str]]]) -> None:
    self.stack.append((self.pos, tag))

  def handle_endtag(self, tag: str) -> None:
    if not self.stack:
      self.msg(f'unmatched closing tag at top level: {tag}')
      return
    if self.stack[-1][1] == tag:
      self.stack.pop()
      return
    for i in reversed(range(len(self.stack))):
      if self.stack[i][1] == tag: # found match.
        self.msg(f'unmatched closing tag: {tag}')
        for p, t in reversed(self.stack[i+1:]):
          self.msg(f'note: ignoring open `{t}` here', pos=p)
        self.msg(f'note: could match here', pos=self.stack[i][0])
        return
    self.msg(f'unmatched closing tag: {tag}')

  def handle_decl(self, decl: str) -> None:
    if decl == 'DOCTYPE html' and self.pos == (1, 1):
      self.found_leading_doctype = True
      return
    self.msg(f'decl: {decl!r}')

  def unknown_decl(self, data: str) -> None:
    self.msg(f'unknown decl: {data!r}')

  def handle_data(self, data: str) -> None: pass

  def handle_pi(self, data: str) -> None:
    self.msg(f'processing instruction: {data!r}')

  @property
  def pos(self) -> Pos:
    line1, col0 = self.getpos()
    return (line1-1, col0)

  def msg(self, msg: str, pos:Pos=None) -> None:
    if pos is None: pos = self.pos
    print(f'{self.path}:{pos[0]+1}:{pos[1]+1}: {msg}')


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.


from pickle import HIGHEST_PROTOCOL, dump as _dump, load as _load
from sys import stdout
from typing import Any, BinaryIO


def load_pickle(file, *, fix_imports=True, encoding="ASCII", errors="strict") -> Any:
  return _load(file, fix_imports=fix_imports, encoding=encoding, errors=errors)


def write_pickle(file:BinaryIO, obj:Any, protocol=HIGHEST_PROTOCOL, fix_imports=True) -> None:
  _dump(obj, file, protocol=protocol, fix_imports=fix_imports)


def out_pickle(obj:Any, protocol=HIGHEST_PROTOCOL, fix_imports=True) -> None:
  _dump(obj, stdout.buffer, protocol=protocol, fix_imports=fix_imports)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Html type hierarchy.
'''

import re
from html import escape as _escape
from typing import (Any, AnyStr as _AnyStr, Callable, Dict, Iterable, Iterator, List, NoReturn, Optional, Tuple, Type, TypeVar,
  Union, cast)

from ..exceptions import ConflictingValues, DeleteNode, FlattenNode, MultipleMatchesError, NoMatchError
from ..markup import Mu, MuAttrs, MuChild, _Mu, _MuChild
from . import semantics


DtDdPair = Tuple[List['Dt'],List['Dd']]


class HtmlNode(Mu):
  'Abstract HTML node; root class of the hierarchy. For the HTML tag, use `Html`.'

  inline_tags = semantics.phrasing_tags
  void_tags = semantics.void_tags
  ws_sensitive_tags = semantics.ws_sensitive_tags
  tag_types:Dict[str,Type[Mu]] = {}

  @classmethod
  def parse(Class:Type[_Mu], source:_AnyStr, **kwargs:Any) -> _Mu:
    from html5_parser import parse
    if 'treebuilder' in kwargs: raise ValueError('HtmlNode.parse() requires default `lxml` treebuilder option.')
    if isinstance(source, bytes): kwargs['transport_encoding'] = 'utf-8'
    etree = parse(source, return_root=True, **kwargs)
    return Class.from_etree(etree)


  def esc_attr_val(self, val:str) -> str: return _escape(val, quote=True)

  def esc_text(self, text:str) -> str: return _escape(text, quote=False)

  @property
  def is_phrasing(self) -> bool:
    'Wether or not the node is phrasing content.'
    return False


  @property
  def attr_urls(self) -> Iterator[str]:
    yield from self.iter_visit(pre=_attr_urls_visit)


def _attr_urls_visit(node:HtmlNode) -> Iterator[str]:
  for k, v in node.attrs.items():
    if k in attr_keys_for_url_vals:
      if k == 'srcset':
        for el in v.split(','):
          src, space, descriptor = el.partition(' ')
          yield src
      else:
        yield v


attr_keys_for_url_vals = frozenset({
  'action', # Form URI.
  'cite',
  'data', # Object.
  'formaction',
  'href',
  'poster',
  'src',
  'srcset',
})


def html_id_for(title:str) -> str:
  '''
  HTML4 IDs consist of ASCII letters, digits, '_', '-' and '.'
  HTML5 no longer has this restriction.
  We choose to restrict IDs to Unicode letters, digits, '_', '-' and '.'
  '''
  return html_id_invalid_re.sub('_', title)

html_id_invalid_re = re.compile(r'[^-.\w]+')


def _tag(Subclass:Type[_Mu]) -> Type[_Mu]:
  'Decorator for associating a concrete subclass with the lowercase tag matching its name.'
  assert issubclass(Subclass, Mu)
  Subclass.tag = Subclass.__name__.lower()
  HtmlNode.tag_types[Subclass.tag] = Subclass
  return Subclass


_Child = TypeVar('_Child', bound=Mu)
_Self = TypeVar('_Self', bound=Mu)

_Accessor = Callable[[_Self],_Child]

def _single(AccesseeClass:Type[_Child]) -> _Accessor:
  'Wrapper function for creating a single-child accessor.'
  tag = AccesseeClass.tag

  def html_single_accessor(self:_Self) -> _Child:
    for c in self.ch:
      if isinstance(c, Mu) and c.tag == tag: return cast(_Child, c)
    return self.append(AccesseeClass())

  return property(html_single_accessor) # type: ignore


# Categories.

class HtmlEmbedded(HtmlNode):
  'Embedded content category.'


class HtmlFlow(HtmlNode):
  'Flow content category.'


class HtmlHeading(HtmlNode):
  'Heading content category.'


class HtmlInteractive(HtmlNode):
  'Interactive content category.'


class HtmlMetadata(HtmlNode):
  'Metadata content category: all node types that are metadata.'


class HtmlPalpable(HtmlNode):
  'Palbable content category.'


class HtmlPhrasing(HtmlNode):
  '''
  Phrasing content category: all node types that are phrasing content.
  Mnemonic: "If it can be inside a sentence, it's phrasing content."
  '''
  @property
  def is_phrasing(self) -> bool: return True # Note: this is overridden for certain special cases.


class HtmlSectioning(HtmlNode):
  'All node types that are sectioning content.'


class HtmlSectioningRoot(HtmlNode):
  'All node types that are sectioning roots.'


# Content models.

class HtmlNoContent(HtmlNode):
  '''
  Html elements that cannot have children.
  Note that the semantics of void elements are partially enforced here,
  and additionally via the set of tags in `.semantics.void_elements`.
  '''

  def append(self, child:_MuChild) -> _MuChild:
    raise TypeError(f'element cannot have child content: {self!r}')


class HtmlHeadingContent(HtmlNode):
  '''
  Heading content model: all node types that can contain heading elements.
  Note: this class represents <em: parents> of heading elements.
  The superclass of H1-H6 heading elements themselves is `Heading`.
  '''

  def h1(self, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None,**kw_attrs:Any) -> 'H1':
    return self.append(H1(attrs=attrs, ch=ch, cl=cl, **kw_attrs))

  def h2(self, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None,**kw_attrs:Any) -> 'H2':
    return self.append(H2(attrs=attrs, ch=ch, cl=cl, **kw_attrs))

  def h3(self, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None,**kw_attrs:Any) -> 'H3':
    return self.append(H3(attrs=attrs, ch=ch, cl=cl, **kw_attrs))

  def h4(self, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None,**kw_attrs:Any) -> 'H4':
    return self.append(H4(attrs=attrs, ch=ch, cl=cl, **kw_attrs))

  def h5(self, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None,**kw_attrs:Any) -> 'H5':
    return self.append(H5(attrs=attrs, ch=ch, cl=cl, **kw_attrs))

  def h6(self, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None,**kw_attrs:Any) -> 'H6':
    return self.append(H6(attrs=attrs, ch=ch, cl=cl, **kw_attrs))

  @property
  def heading(self) -> Optional[HtmlNode]:
    for el in self.ch:
      if isinstance(el, Heading): return el
    return None


class HtmlMetadataContent(HtmlNode):
  'Metadata content model: all node types that can contain metadata.'
  # TODO: metadata constructor methods.


class HtmlPhrasingContent(HtmlNode):
  '''
  Phrasing content model: all node types that can contain phrasing content.
  '''
  # TODO: all phrasing constructor methods.


class HtmlScriptSupporting(HtmlNode):

  def script(self, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None,**kw_attrs:Any) -> 'Script':
    return self.append(Script(attrs=attrs, ch=ch, cl=cl, **kw_attrs))


class HtmlTextContent(HtmlNode):
  'All node types that can contain text content.'


class HtmlTransparentContent(HtmlNode):
  'Transparent content elements.'


class HtmlTransparentPhrasing(HtmlTransparentContent):
  'Nodes are considered phrasing content if all of their children are phrasing content.'
  @property
  def is_phrasing(self) -> bool:
    return all((not isinstance(c, Html) or c.is_phrasing) for c in self.ch)


class HtmlFlowContent(HtmlHeadingContent, HtmlPhrasingContent):
  '''
  All elements that can contain flow content.
  '''
  # TODO: flow constructor methods.


# Elements.


@_tag
class A(HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing, HtmlTransparentContent):
  '''
  Creates a hyperlink to other web pages, files, locations within the same page, email addresses, or any other URL.

  Categories:
    Interactive: if the element has an href attribute.

  Content model:
    Transparent: there must be no interactive content or a element descendants.

  Contexts for use: Phrasing.
  '''


@_tag
class Abbr(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents an abbreviation or acronym; the optional title attribute can provide an expansion or description for the abbreviation.

  Contexts for use: Phrasing.
  '''


@_tag
class Address(HtmlFlow, HtmlPalpable, HtmlFlowContent):
  '''
  Indicates that the enclosed HTML provides contact information for a person or people, or for an organization.

  Content model:
    Flow: with no heading content descendants, no sectioning content descendants, and no header, footer, or address element descendants.

  Contexts for use: Flow.
  '''


@_tag
class Area(HtmlFlow, HtmlPhrasing, HtmlNoContent):
  '''
  Defines a hot-spot region on an image, and optionally associates it with a hypertext link. This is used only within a <map> element.

  Contexts for use: Where phrasing content is expected, but only if there is a map element ancestor.
  '''


@_tag
class Article(HtmlFlow, HtmlPalpable, HtmlSectioning, HtmlFlowContent):
  '''
  Represents a self-contained composition in a document, page, application, or site, which is intended to be independently distributable or reusable.

  Contexts for use: Flow.
  '''


@_tag
class Aside(HtmlFlow, HtmlPalpable, HtmlSectioning, HtmlFlowContent):
  '''
  Represents a portion of a document whose content is only indirectly related to the document's main content.

  Contexts for use: Flow.
  '''


@_tag
class Audio(HtmlEmbedded, HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing):
  '''
  Used to embed sound content in documents. It may contain one or more audio sources, represented using the src attribute or the <source> element: the browser will choose the most suitable one. It can also be the destination for streamed media, using a MediaStream.

  Categories:
    Interactive: if the element has a controls attribute.
    Palpable: if the element has a controls attribute.

  Content model:
    Zero or more source elements, then zero or more track elements, then transparent, but with no media element descendants: if the element does not have a src attribute.
    Zero or more track elements, then transparent, but with no media element descendants: if the element has a src attribute.

  Contexts for use: Embedded.
  '''


@_tag
class B(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Used to draw the reader's attention to the element's contents, which are not otherwise granted special importance.

  Contexts for use: Phrasing.
  '''


@_tag
class Base(HtmlMetadata, HtmlNoContent):
  '''
  Specifies the base URL to use for all relative URLs contained within a document. There can be only one <base> element in a document.

  Contexts for use: In a head element containing no other base elements.
  '''


@_tag
class Bdi(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Tells the browser's bidirectional algorithm to treat the text it contains in isolation from its surrounding text.

  Contexts for use: Phrasing.
  '''


@_tag
class Bdo(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Overrides the current directionality of text, so that the text within is rendered in a different direction.

  Contexts for use: Phrasing.
  '''


@_tag
class Blockquote(HtmlFlow, HtmlPalpable, HtmlSectioningRoot, HtmlFlowContent):
  '''
  Indicates that the enclosed text is an extended quotation. Usually, this is rendered visually by indentation (see Notes for how to change it). A URL for the source of the quotation may be given using the cite attribute, while a text representation of the source can be given using the <cite> element.

  Contexts for use: Flow.
  '''


@_tag
class Body(HtmlSectioningRoot, HtmlFlowContent):
  '''
  Represents the content of an HTML document. There can be only one <body> element in a document.

  Contexts for use: As the second element in an html element.
  '''

  @property
  def header(self) -> 'Header': return self._single(Header)

  @property
  def nav(self) -> 'Nav': return self._single(Nav)


  @property
  def main(self) -> 'Main': return self._single(Main)


  @property
  def footer(self) -> 'Footer': return self._single(Footer)


@_tag
class Br(HtmlFlow, HtmlPhrasing, HtmlNoContent):
  '''
  Produces a line break in text (carriage-return). It is useful for writing a poem or an address, where the division of lines is significant.

  Contexts for use: Phrasing.
  '''


@_tag
class Button(HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents a clickable button, which can be used in forms or anywhere in a document that needs simple, standard button functionality.

  Categories:
    Listed, labelable, submittable, and autocapitalize-inheriting form-associated element: None.

  Content model:
    Phrasing: there must be no interactive content descendant.

  Contexts for use: Phrasing.
  '''


@_tag
class Canvas(HtmlEmbedded, HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlTransparentContent):
  '''
  Used with either the canvas scripting API or the WebGL API to draw graphics and animations.

  Content model:
    Transparent: with no interactive content descendants except for a elements, img elements with usemap attributes, button elements, input elements whose type attribute are in the Checkbox or Radio Button states, input elements that are buttons, select elements with a multiple attribute or a display size greater than 1, and elements that would not be interactive content except for having the tabindex attribute specified.

  Contexts for use: Embedded.
  '''


@_tag
class Caption(HtmlFlowContent):
  '''
  Specifies the caption (or title) of a table, and if used is always the first child of a <table>.

  Content model:
    Flow: with no descendant table elements.

  Contexts for use: As the first element child of a table element.
  '''


@_tag
class Cite(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Used to describe a reference to a cited creative work, and must include the title of that work.

  Contexts for use: Phrasing.
  '''


@_tag
class Code(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Displays its contents styled in a fashion intended to indicate that the text is a short fragment of computer code.

  Contexts for use: Phrasing.
  '''


@_tag
class Col(HtmlNoContent):
  '''
  Defines a column within a table and is used for defining common semantics on all common cells. It is generally found within a <colgroup> element.

  Contexts for use: As a child of a colgroup element that doesn't have a span attribute.
  '''


@_tag
class Colgroup(HtmlNode):
  '''
  Defines a group of columns within a table.

  Content model:
    No: if the span attribute is present.
    Zero or more col and template elements: if the span attribute is absent.

  Contexts for use: As a child of a table element, after any caption elements and before any thead, tbody, tfoot, and tr elements.
  '''


@_tag
class Data(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Links a given content with a machine-readable translation. If the content is time- or date-related, the <time> element must be used.

  Contexts for use: Phrasing.
  '''


@_tag
class Datalist(HtmlFlow, HtmlPhrasing):
  '''
  Contains a set of <option> elements that represent the values available for other controls.

  Content model:
    Either: phrasing: None.
    Or: Zero or more option and script-supporting elements: None.

  Contexts for use: Phrasing.
  '''


@_tag
class Dd(HtmlFlowContent):
  '''
  Provides the details about or the definition of the preceding term (<dt>) in a description list (<dl>).

  Contexts for use: After dt or dd elements inside div elements that are children of a dl element, After dt or dd elements inside dl elements.
  '''


@_tag
class Del(HtmlFlow, HtmlPhrasing, HtmlTransparentContent):
  '''
  Represents a range of text that has been deleted from a document.

  Contexts for use: Phrasing.
  '''


@_tag
class Details(HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlSectioningRoot):
  '''
  Creates a disclosure widget in which information is visible only when the widget is toggled into an "open" state.

  Content model:
    One summary element followed by flow: None.

  Contexts for use: Flow.
  '''
  # TODO: summary property.


@_tag
class Dfn(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Used to indicate the term being defined within the context of a definition phrase or sentence.

  Content model:
    Phrasing: there must be no dfn element descendants.

  Contexts for use: Phrasing.
  '''

  def dfn(self) -> NoReturn: raise TypeError('`dfn` elements cannot have `dfn` descendants.')


@_tag
class Dialog(HtmlFlow, HtmlSectioningRoot, HtmlFlowContent):
  '''
  Represents a dialog box or other interactive component, such as an inspector or window.

  Contexts for use: Flow.
  '''


@_tag
class Div(HtmlFlow, HtmlPalpable, HtmlFlowContent):
  '''
  The generic container for flow content. It has no effect on the content or layout until styled using CSS.

  Content model:
    Flow: if the element is not a child of a dl element.
    One or more dt elements followed by one or more dd elements, optionally intermixed with script-supporting elements: if the element is a child of a dl element.

  Contexts for use: As a child of a dl element, Flow.
  '''


@_tag
class Dl(HtmlFlow, HtmlPalpable):
  '''
  Represents a description list. The element encloses a list of groups of terms (specified using the <dt> element) and descriptions (provided by <dd> elements). Common uses for this element are to implement a glossary or to display metadata (a list of key-value pairs).

  Categories:
    Palpable: if the element's children include at least one name-value group.

  Content model:
    Either: Zero or more groups each consisting of one or more dt elements followed by one or more dd elements, optionally intermixed with script-supporting elements: None.
    Or: One or more div elements, optionally intermixed with script-supporting elements: None.

  Contexts for use: Flow.
  '''

  @property
  def dt_dd_groups(self) -> Iterable[Union[Div,DtDdPair]]:
    pair:Optional[DtDdPair] = None
    for c in self.child_nodes():
      if isinstance(c, Div):
        yield c
      elif isinstance(c, Dt):
        if pair:
          if pair[1]: # Yield previous pair and start new one.
            yield pair
            pair = ([c], [])
          else: # Accumulate additional Dt.
            pair[0].append(c)
            continue
        else:
          pair = ([c], [])
      elif isinstance(c, Dd):
        if pair:
          pair[1].append(c) # Accumulate additional Dt.
        else: # This case is invalid, but we can create a pair with no leading Dt.
          pair = ([], [c])
    if pair: yield pair


@_tag
class Dt(HtmlFlowContent):
  '''
  Specifies a term in a description list, and as such must be used inside a <dl> element.

  Content model:
    Flow: with no header, footer, sectioning content, or heading content descendants.

  Contexts for use: Before dd or dt elements inside div elements that are children of a dl element, Before dd or dt elements inside dl elements.
  '''


@_tag
class Em(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Marks text that has stress emphasis. The <em> element can be nested, with each level of nesting indicating a greater degree of emphasis.

  Contexts for use: Phrasing.
  '''


@_tag
class Embed(HtmlEmbedded, HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing, HtmlNoContent):
  '''
  Embeds external content at the specified point in the document. This content is provided by an external application or other source of interactive content such as a browser plug-in.

  Contexts for use: Embedded.
  '''


@_tag
class Fieldset(HtmlFlow, HtmlPalpable, HtmlSectioningRoot):
  '''
  Used to group several controls as well as labels (<label>) within a web form.

  Categories:
    Listed and autocapitalize-inheriting form-associated element: None.

  Content model:
    Optionally a legend element, followed by flow: None.

  Contexts for use: Flow.
  '''


@_tag
class Figcaption(HtmlFlowContent):
  '''
  Represents a caption or legend describing the rest of the contents of its parent <figure> element.

  Contexts for use: As the first or last child of a figure element.
  '''


@_tag
class Figure(HtmlFlow, HtmlPalpable, HtmlSectioningRoot):
  '''
  Represents self-contained content, potentially with an optional caption, which is specified using the (<figcaption>) element.

  Content model:
    Either: one figcaption element followed by flow: None.
    Or: flow: None.
    Or: flow content followed by one figcaption element: None.

  Contexts for use: Flow.
  '''


@_tag
class Footer(HtmlFlow, HtmlPalpable, HtmlFlowContent):
  '''
  Represents a footer for its nearest sectioning content or sectioning root element. A footer typically contains information about the author of the section, copyright data or links to related documents.

  Content model:
    Flow: with no header or footer element descendants.

  Contexts for use: Flow.
  '''


@_tag
class Form(HtmlFlow, HtmlPalpable, HtmlFlowContent):
  '''
  Represents a document section that contains interactive controls for submitting information to a web server.

  Content model:
    Flow: with no form element descendants.

  Contexts for use: Flow.
  '''


class Heading(HtmlPhrasingContent):
  '''
  Parent class for H1-H6 heading elements, which represent six levels of section headings. <h1> is the highest section level and <h6> is the lowest.

  Contexts for use: As a child of an hgroup element, Flow.
  '''

  @classmethod
  def for_level(cls, level:int, *, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None, **kw_attrs:Any) -> 'Heading':
    c = _heading_classes[min(level, 6) - 1]
    return c(attrs=attrs, ch=ch, cl=cl, **kw_attrs)


@_tag
class H1(Heading):
  'H1 heading.'

@_tag
class H2(Heading):
  'H2 heading.'

@_tag
class H3(Heading):
  'H3 heading.'

@_tag
class H4(Heading):
  'H4 heding.'

@_tag
class H5(Heading):
  'H5 heading.'

@_tag
class H6(Heading):
  'H6 heading.'

_heading_classes:List[Type[Heading]] = [H1, H2, H3, H4, H5, H6]


@_tag
class Head(HtmlMetadataContent):
  '''
  Contains machine-readable information (metadata) about the document, like its title, scripts, and style sheets.

  Content model:
    Zero or more elements of metadata content, of which no more than one is a title element and no more than one is a base element: if the document is an iframe srcdoc document or if title information is available from a higher-level protocol.
    One or more elements of metadata content, of which exactly one is a title element and no more than one is a base element: Otherwise.

  Contexts for use: As the first element in an html element.
  '''

  @property
  def title(self) -> 'Title': return self._single(Title)

  def add_stylesheet(self, url:str, media='all') -> None:
    self.append(Link(rel='stylesheet', type='text/css', media=media, href=url))

  @property
  def style(self) -> 'Style': return self._single(Style)


@_tag
class Header(HtmlFlow, HtmlPalpable, HtmlFlowContent):
  '''
  Represents introductory content, typically a group of introductory or navigational aids. It may contain some heading elements but also a logo, a search form, an author name, and other elements.

  Content model:
    Flow: with no header or footer element descendants.

  Contexts for use: Flow.
  '''


@_tag
class Hgroup(HtmlFlow, HtmlHeading, HtmlPalpable):
  '''
  Represents a multi-level heading for a section of a document. It groups a set of <h1>–<h6> elements.

  Content model:
    One or more h1, h2, h3, h4, h5, h6 elements, optionally intermixed with script-supporting elements: None.

  Contexts for use: Flow.
  '''


@_tag
class Hr(HtmlFlow, HtmlNoContent):
  '''
  Represents a thematic break between paragraph-level elements: for example, a change of scene in a story, or a shift of topic within a section.

  Contexts for use: Flow.
  '''


@_tag
class Html(HtmlNode):
  '''
  Represents the root (top-level element) of an HTML document, so it is also referred to as the root element. All other elements must be descendants of this element.

  Content model:
    A head element followed by a body element: None.

  Contexts for use: As document's document element, Wherever a subdocument fragment is allowed in a compound document.
  '''

  def render(self, newline:bool=True) -> Iterator[str]:
    yield '<!DOCTYPE html>\n'
    yield from super().render(newline=newline)

  @property
  def body(self) -> Body: return self._single(Body)

  @property
  def head(self) -> Head: return self._single(Head)

  @staticmethod
  def doc(*, title:str, charset='utf-8', lang='en') -> 'Html':
    html = Html(lang=lang)
    head = html.head
    head.title.append(title)
    head.append(Meta(charset=charset))
    return html


@_tag
class I(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents a range of text that is set off from the normal text for some reason. Some examples include technical terms, foreign language phrases, or fictional character thoughts. It is typically displayed in italic type.

  Contexts for use: Phrasing.
  '''


@_tag
class Iframe(HtmlEmbedded, HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing, HtmlNoContent):
  '''
  Represents a nested browsing context, embedding another HTML page into the current one.

  Contexts for use: Embedded.
  '''


@_tag
class Img(HtmlEmbedded, HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing, HtmlNoContent):
  '''
  Embeds an image into the document. It is a replaced element.

  Categories:
    Form-associated element: None.
    Interactive: if the element has a usemap attribute.

  Contexts for use: Embedded.
  '''


@_tag
class Input(HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing, HtmlNoContent):
  '''
  Used to create interactive controls for web-based forms in order to accept data from the user; a wide variety of types of input data and control widgets are available, depending on the device and user agent.

  Categories:
    Interactive: if the type attribute is not in the hidden state.
    Listed, labelable, submittable, resettable, and autocapitalize-inheriting form-associated element: if the type attribute is not in the hidden state.
    Listed, submittable, resettable, and autocapitalize-inheriting form-associated element: if the type attribute is in the hidden state.
    Palpable: if the type attribute is not in the hidden state.

  Contexts for use: Phrasing.
  '''


@_tag
class Ins(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlTransparentContent):
  '''
  Represents a range of text that has been added to a document.

  Contexts for use: Phrasing.
  '''


@_tag
class Kbd(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents a span of inline text denoting textual user input from a keyboard, voice input, or any other text entry device.

  Contexts for use: Phrasing.
  '''


@_tag
class Label(HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents a caption for an item in a user interface.

  Content model:
    Phrasing: with no descendant labelable elements unless it is the element's labeled control, and no descendant label elements.

  Contexts for use: Phrasing.
  '''


@_tag
class Legend(HtmlPhrasingContent):
  '''
  Represents a caption for the content of its parent <fieldset>.

  Contexts for use: As the first child of a fieldset element.
  '''


@_tag
class Li(HtmlFlowContent):
  '''
  Used to represent an item in a list.

  Contexts for use: Inside menu elements, Inside ol elements, Inside ul elements.
  '''


@_tag
class Link(HtmlFlow, HtmlMetadata, HtmlPhrasing, HtmlNoContent):
  '''
  The HTML External Resource Link element (<link>) specifies relationships between the current document and an external resource. This element is most commonly used to link to stylesheets, but is also used to establish site icons (both "favicon" style icons and icons for the home screen and apps on mobile devices) among other things.

  Categories:
    Flow: if the element is allowed in the body.
    Phrasing: if the element is allowed in the body.

  Contexts for use: If the element is allowed in the body: where phrasing content is expected, In a noscript element that is a child of a head element, Where metadata content is expected.
  '''


@_tag
class Main(HtmlFlow, HtmlPalpable, HtmlFlowContent):
  '''
  Represents the dominant content of the <body> of a document. The main content area consists of content that is directly related to or expands upon the central topic of a document, or the central functionality of an application.

  Contexts for use: Where flow content is expected, but only if it is a hierarchically correct main element.
  '''


@_tag
class Map(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlTransparentContent):
  '''
  Used with <area> elements to define an image map (a clickable link area).

  Contexts for use: Phrasing.
  '''


@_tag
class Mark(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents text which is marked or highlighted for reference or notation purposes, due to the marked passage's relevance or importance in the enclosing context.

  Contexts for use: Phrasing.
  '''


@_tag
class Math(HtmlNode):
  '''
  Represents a mathematical expression.

  Contexts for use: Phrasing.
  '''


@_tag
class Menu(HtmlFlow, HtmlPalpable):
  '''
  Represents a group of commands that a user can perform or activate. This includes both list menus, which might appear across the top of a screen, as well as context menus, such as those that might appear underneath a button after it has been clicked.

  Categories:
    Palpable: if the element's children include at least one li element.

  Content model:
    Zero or more li and script-supporting elements: None.

  Contexts for use: Flow.
  '''


@_tag
class Meta(HtmlFlow, HtmlMetadata, HtmlPhrasing, HtmlNoContent):
  '''
  Represents metadata that cannot be represented by other HTML meta-related elements, like <base>, <link>, <script>, <style> or <title>.

  Categories:
    Flow: if the itemprop attribute is present.
    Phrasing: if the itemprop attribute is present.

  Contexts for use: If the charset attribute is present, or if the element's http-equiv attribute is in the Encoding declaration state: in a head element, If the http-equiv attribute is present but not in the Encoding declaration state: in a head element, If the http-equiv attribute is present but not in the Encoding declaration state: in a noscript element that is a child of a head element, If the itemprop attribute is present: where metadata content is expected, If the itemprop attribute is present: where phrasing content is expected, If the name attribute is present: where metadata content is expected.
  '''


@_tag
class Meter(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents either a scalar value within a known range or a fractional value.

  Categories:
    Labelable element: None.

  Content model:
    Phrasing: there must be no meter element descendants.

  Contexts for use: Phrasing.
  '''


@_tag
class Nav(HtmlFlow, HtmlPalpable, HtmlSectioning, HtmlFlowContent):
  '''
  Represents a section of a page whose purpose is to provide navigation links, either within the current document or to other documents. Common examples of navigation sections are menus, tables of contents, and indexes.

  Contexts for use: Flow.
  '''


@_tag
class Noscript(HtmlFlow, HtmlMetadata, HtmlPhrasing):
  '''
  Defines a section of HTML to be inserted if a script type on the page is unsupported or if scripting is currently turned off in the browser.

  Content model:
    Otherwise: text that conforms to the requirements given in the prose: None.
    When scripting is disabled, in a head element: in any order, zero or more link elements, zero or more style elements, and zero or more meta elements: None.
    When scripting is disabled, not in a head element: transparent: there must be no noscript element descendants.

  Contexts for use: In a head element of an HTML document, if there are no ancestor noscript elements, Where phrasing content is expected in HTML documents, if there are no ancestor noscript elements.
  '''


@_tag
class Object(HtmlEmbedded, HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing):
  '''
  Represents an external resource, which can be treated as an image, a nested browsing context, or a resource to be handled by a plugin.

  Categories:
    Interactive: if the element has a usemap attribute.
    Listed and submittable form-associated element: None.

  Content model:
    Zero or more param elements, then, transparent: None.

  Contexts for use: Embedded.
  '''


@_tag
class Ol(HtmlFlow, HtmlPalpable):
  '''
  Represents an ordered list of items, typically rendered as a numbered list.

  Categories:
    Palpable: if the element's children include at least one li element.

  Content model:
    Zero or more li and script-supporting elements: None.

  Contexts for use: Flow.
  '''


@_tag
class Optgroup(HtmlNode):
  '''
  Creates a grouping of options within a <select> element.

  Content model:
    Zero or more option and script-supporting elements: None.

  Contexts for use: As a child of a select element.
  '''


@_tag
class Option(HtmlTextContent):
  '''
  Used to define an item contained in a <select>, an <optgroup>, or a <datalist> element. As such, <option> can represent menu items in popups and other lists of items in an HTML document.

  Content model:
    No: if the element has a label attribute and a value attribute.
    Text: if the element has a label attribute but no value attribute.
    Text: if the element has no label attribute and is a child of a datalist element.
    Text that is not inter-element whitespace: if the element has no label attribute and is not a child of a datalist element.

  Contexts for use: As a child of a datalist element, As a child of a select element, As a child of an optgroup element.
  '''


@_tag
class Output(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Used to inject the results of a calculation or the outcome of a user action.

  Categories:
    Listed, labelable, resettable, and autocapitalize-inheriting form-associated element: None.

  Contexts for use: Phrasing.
  '''


@_tag
class P(HtmlFlow, HtmlPalpable, HtmlPhrasingContent):
  '''
  Represents a paragraph.

  Contexts for use: Flow.
  '''


@_tag
class Param(HtmlNoContent):
  '''
  Defines parameters for an <object> element.

  Contexts for use: As a child of an object element, before any flow content.
  '''


@_tag
class Picture(HtmlEmbedded, HtmlFlow, HtmlPhrasing):
  '''
  Contains zero or more <source> elements and one <img> element to provide versions of an image for different display/device scenarios.

  Content model:
    Zero or more source elements, followed by one img element, optionally intermixed with script-supporting elements: None.

  Contexts for use: Embedded.
  '''


@_tag
class Pre(HtmlFlow, HtmlPalpable, HtmlPhrasingContent):
  '''
  Represents preformatted text which is to be presented exactly as written in the HTML file.

  Contexts for use: Flow.
  '''


@_tag
class Progress(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Displays an indicator showing the completion progress of a task, typically displayed as a progress bar.

  Categories:
    Labelable element: None.

  Content model:
    Phrasing: there must be no progress element descendants.

  Contexts for use: Phrasing.
  '''


@_tag
class Q(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Indicates that the enclosed text is a short inline quotation. Most modern browsers implement this by surrounding the text in quotation marks.

  Contexts for use: Phrasing.
  '''


@_tag
class Rb(HtmlTextContent):
  '''
  Used to delimit the base text component of a <ruby> annotation, i.e. the text that is being annotated.

  Contexts for use: As a child of a ruby element.
  '''


@_tag
class Rp(HtmlTextContent):
  '''
  Used to provide fall-back parentheses for browsers that do not support display of ruby annotations using the <ruby> element.

  Contexts for use: As a child of a ruby element, either immediately before or immediately after an rt element.
  '''


@_tag
class Rt(HtmlPhrasingContent):
  '''
  Specifies the ruby text component of a ruby annotation, which is used to provide pronunciation, translation, or transliteration information for East Asian typography. The <rt> element must always be contained within a <ruby> element.

  Contexts for use: As a child of a ruby element.
  '''


@_tag
class Rtc(HtmlPhrasingContent):
  '''
  Embraces semantic annotations of characters presented in a ruby of <rb> elements used inside of <ruby> element. <rb> elements can have both pronunciation (<rt>) and semantic (<rtc>) annotations.

  Contexts for use: As a child of a ruby element.
  '''


@_tag
class Ruby(HtmlFlow, HtmlPalpable, HtmlPhrasing):
  '''
  Represents a ruby annotation. Ruby annotations are for showing pronunciation of East Asian characters.

  Content model:
    See prose: None.

  Contexts for use: Phrasing.
  '''


@_tag
class S(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Renders text with a strikethrough, or a line through it. Use the <s> element to represent things that are no longer relevant or no longer accurate. However, <s> is not appropriate when indicating document edits; for that, use the <del> and <ins> elements, as appropriate.

  Contexts for use: Phrasing.
  '''


@_tag
class Samp(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Used to enclose inline text which represents sample (or quoted) output from a computer program.

  Contexts for use: Phrasing.
  '''


@_tag
class Script(HtmlFlow, HtmlMetadata, HtmlPhrasing):
  '''
  Used to embed or reference executable code; this is typically used to embed or refer to JavaScript code.

  Categories:
    Script-supporting element: None.

  Content model:
    : if there is a src attribute, the element must be either empty or contain only script documentation that also matches script content restrictions.
    : if there is no src attribute, depends on the value of the type attribute, but must match script content restrictions.

  Contexts for use: Phrasing, Where metadata content is expected, Where script-supporting elements are expected.
  '''


@_tag
class Section(HtmlFlow, HtmlPalpable, HtmlSectioning, HtmlFlowContent):
  '''
  Represents a standalone section which does not have a more specific semantic element to represent it.

  Contexts for use: Flow.
  '''


@_tag
class Select(HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing):
  '''
  Represents a control that provides a menu of options

  Categories:
    Listed, labelable, submittable, resettable, and autocapitalize-inheriting form-associated element: None.

  Content model:
    Zero or more option, optgroup, and script-supporting elements: None.

  Contexts for use: Phrasing.
  '''


@_tag
class Slot(HtmlFlow, HtmlPhrasing, HtmlTransparentContent):
  '''
  Used as a placeholder inside a web component that you can fill with your own markup, which lets you create separate DOM trees and present them together.

  Contexts for use: Phrasing.
  '''


@_tag
class Small(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Makes the text font size one size smaller (for example, from large to medium, or from small to x-small) down to the browser's minimum font size.  In HTML5, this element is repurposed to represent side-comments and small print, including copyright and legal text, independent of its styled presentation.

  Contexts for use: Phrasing.
  '''


@_tag
class Source(HtmlNoContent):
  '''
  Specifies multiple media resources for the <picture>, the <audio> element, or the <video> element.

  Contexts for use: As a child of a media element, before any flow content or track elements, As a child of a picture element, before the img element.
  '''


@_tag
class Span(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  A generic inline container for phrasing content, which does not inherently represent anything. It can be used to group elements for styling purposes (using the class or id attributes), or because they share attribute values, such as lang.

  Contexts for use: Phrasing.
  '''


@_tag
class Strong(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Indicates that its contents have strong importance, seriousness, or urgency. Browsers typically render the contents in bold type.

  Contexts for use: Phrasing.
  '''


@_tag
class Style(HtmlMetadata):
  '''
  Contains style information for a document, or part of a document.

  Content model:
    Text that gives a conformant style sheet: None.

  Contexts for use: In a noscript element that is a child of a head element, Where metadata content is expected.
  '''


@_tag
class Sub(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Specifies inline text which should be displayed as subscript for solely typographical reasons.

  Contexts for use: Phrasing.
  '''


@_tag
class Summary(HtmlPhrasing):
  '''
  Specifies a summary, caption, or legend for a <details> element's disclosure box.

  Content model:
    Either: phrasing: None.
    Or: one element of heading: None.

  Contexts for use: As the first child of a details element.
  '''


@_tag
class Sup(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Specifies inline text which is to be displayed as superscript for solely typographical reasons.

  Contexts for use: Phrasing.
  '''


@_tag
class Svg(HtmlNode):
  '''
  An SVG vector graphic.

  Contexts for use: Phrasing.
  '''


@_tag
class Table(HtmlFlow, HtmlPalpable):
  '''
  Represents tabular data: information presented in a two-dimensional table comprised of rows and columns of cells containing data.

  Content model:
    In this order: optionally a caption element, followed by zero or more colgroup elements, followed optionally by a thead element, followed by either zero or more tbody elements or one or more tr elements, followed optionally by a tfoot element, optionally intermixed with one or more script-supporting elements: None.

  Contexts for use: Flow.
  '''
  # TODO: def caption



@_tag
class Tbody(HtmlNode):
  '''
  Encapsulates a set of table rows (<tr> elements), indicating that they comprise the body of the table (<table>).

  Content model:
    Zero or more tr and script-supporting elements: None.

  Contexts for use: As a child of a table element, after any caption, colgroup, and thead elements, but only if there are no tr elements that are children of the table element.
  '''


@_tag
class Td(HtmlSectioningRoot, HtmlFlowContent):
  '''
  Defines a cell of a table that contains data. It participates in the table model.

  Contexts for use: As a child of a tr element.
  '''


@_tag
class Template(HtmlFlow, HtmlMetadata, HtmlPhrasing):
  '''
  Used to hold HTML that is not to be rendered immediately when a page is loaded but may be instantiated subsequently during runtime using JavaScript.

  Categories:
    Script-supporting element: None.

  Content model:
    Nothing (for clarification, see example): None.

  Contexts for use: As a child of a colgroup element that doesn't have a span attribute, Phrasing, Where metadata content is expected, Where script-supporting elements are expected.
  '''


@_tag
class Textarea(HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing, HtmlTextContent):
  '''
  Represents a multi-line plain-text editing control, useful when you want to allow users to enter a sizeable amount of free-form text, for example a comment on a review or feedback form.

  Categories:
    Listed, labelable, submittable, resettable, and autocapitalize-inheriting form-associated element: None.

  Contexts for use: Phrasing.
  '''


@_tag
class Tfoot(HtmlNode):
  '''
  Defines a set of rows summarizing the columns of the table.

  Content model:
    Zero or more tr and script-supporting elements: None.

  Contexts for use: As a child of a table element, after any caption, colgroup, thead, tbody, and tr elements, but only if there are no other tfoot elements that are children of the table element.
  '''


@_tag
class Th(HtmlFlowContent):
  '''
  Defines a cell as header of a group of table cells. The exact nature of this group is defined by the scope and headers attributes.

  Content model:
    Flow: with no header, footer, sectioning content, or heading content descendants.

  Contexts for use: As a child of a tr element.
  '''


@_tag
class Thead(HtmlNode):
  '''
  Defines a set of rows defining the head of the columns of the table.

  Content model:
    Zero or more tr and script-supporting elements: None.

  Contexts for use: As a child of a table element, after any caption, and colgroup elements and before any tbody, tfoot, and tr elements, but only if there are no other thead elements that are children of the table element.
  '''


@_tag
class Time(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents a specific period in time.

  Content model:
    Otherwise: Text: must match requirements described in prose below.
    Phrasing: if the element has a datetime attribute.

  Contexts for use: Phrasing.
  '''


@_tag
class Title(HtmlMetadata):
  '''
  Defines the document's title that is shown in a browser's title bar or a page's tab.

  Content model:
    Text that is not inter-element whitespace: None.

  Contexts for use: In a head element containing no other title elements.
  '''


@_tag
class Tr(HtmlNode):
  '''
  Defines a row of cells in a table. The row's cells can then be established using a mix of <td> (data cell) and <th> (header cell) elements.

  Content model:
    Zero or more td, th, and script-supporting elements: None.

  Contexts for use: As a child of a table element, after any caption, colgroup, and thead elements, but only if there are no tbody elements that are children of the table element, As a child of a tbody element, As a child of a tfoot element, As a child of a thead element.
  '''


@_tag
class Track(HtmlNoContent):
  '''
  Used as a child of the media elements <audio> and <video>. It lets you specify timed text tracks (or time-based data), for example to automatically handle subtitles. The tracks are formatted in WebVTT format (.vtt files) — Web Video Text Tracks or Timed Text Markup Language (TTML).

  Contexts for use: As a child of a media element, before any flow content.
  '''


@_tag
class U(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents a span of inline text which should be rendered in a way that indicates that it has a non-textual annotation.

  Contexts for use: Phrasing.
  '''


@_tag
class Ul(HtmlFlow, HtmlPalpable):
  '''
  Represents an unordered list of items, typically rendered as a bulleted list.

  Categories:
    Palpable: if the element's children include at least one li element.

  Content model:
    Zero or more li and script-supporting elements: None.

  Contexts for use: Flow.
  '''


@_tag
class Var(HtmlFlow, HtmlPalpable, HtmlPhrasing, HtmlPhrasingContent):
  '''
  Represents the name of a variable in a mathematical expression or a programming context.

  Contexts for use: Phrasing.
  '''


@_tag
class Video(HtmlEmbedded, HtmlFlow, HtmlInteractive, HtmlPalpable, HtmlPhrasing):
  '''
  Embeds a media player which supports video playback into the document.

  Categories:
    Interactive: if the element has a controls attribute.

  Content model:
    Zero or more source elements, then zero or more track elements, then transparent, but with no media element descendants: if the element does not have a src attribute.
    Zero or more track elements, then transparent, but with no media element descendants: if the element has a src attribute.

  Contexts for use: Embedded.
  '''


@_tag
class Wbr(HtmlFlow, HtmlPhrasing, HtmlNoContent):
  '''
  Represents a word break opportunity: a position within text where the browser may optionally break a line, even if its line-breaking rules would not otherwise create a break at that location.

  Contexts for use: Phrasing.
  '''



class InlineStyle(Dict[str,Any]):

  def __str__(self) -> str:
    return ';'.join(f'{k.replace("_", "-")}:{v}' for (k, v) in self.items())



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import inspect
from functools import _find_impl, wraps  # type: ignore
from typing import Any, Callable, DefaultDict, Dict, Optional, Tuple, TypeVar, cast

from .default import Default


class DispatchKeyError(KeyError): pass
class DispatchTypeError(TypeError): pass


# module_name -> method_name -> arg_type.
_dispatched_registries = DefaultDict[str, DefaultDict[str, Dict[type, Callable]]](lambda: DefaultDict(dict))

_Method = TypeVar('_Method', bound=Callable)

def dispatched(method:_Method) -> _Method:
  '''
  Decorator for instance methods to dispatch on first arg (after self).
  Uses the same MRO resolution algorithm as functools.singledispatch.

  Usage:
  | class C():
  |
  |  @dispatched
  |  def describe(self, item:int) -> None: print(f'int: {item}')
  |
  |  @dispatched
  |  def describe(self, item:str) -> None: print(f'str: {item}')

  TODO: poison registration mechanism after first call to prevent subtle MRO/caching errors.
  TODO: detect method vs plain function?
  '''
  assert method.__closure__ is None # type: ignore # Method cannot be a closure.
  registry = _dispatched_registries[method.__module__][method.__name__]
  sig = inspect.signature(method)
  pars = list(sig.parameters.values())
  par_self = pars[0]
  par_arg = pars[1]
  assert par_self.name == 'self'
  t = par_arg.annotation
  if not isinstance(t, type):
    raise TypeError(f'`dispatched` requires type annotation to be a runtime type: {t}') # TODO: allow generics by using __origin__.
  if t in registry:
    a = _source_loc(registry[t])
    b = _source_loc(method)
    raise DispatchTypeError(f'`dispatched` methods collide on argument type:\n{a}\n{b}')
  registry[t] = method

  dispatch_cache: Dict[type, Callable] = {}
  @wraps(method)
  def dispatch(self, arg, *args, **kwargs):
    method = _dispatch(type(arg), registry, dispatch_cache)
    return method(self, arg, *args, **kwargs)

  return cast(_Method, dispatch)


def _source_loc(function: Callable) -> str:
  code = function.__code__
  return f'{code.co_filename}:{code.co_firstlineno}'


def _dispatch(cls: type, registry: Dict[type, Callable], dispatch_cache: Dict[type, Callable]) -> Callable:
  '''
  Dispatch using functools.singledispatch algorithm.
  '''
  try: return dispatch_cache[cls]
  except KeyError: pass
  try:
    m = dispatch_cache[cls] = registry[cls]
    return m
  except KeyError: pass
  try: m = _find_impl(cls, registry)
  except RuntimeError as e: raise DispatchTypeError(cls) from e
  if m is None: raise DispatchTypeError(cls)
  dispatch_cache[cls] = m
  return m


# module_name -> method_name -> (dispatcher_method, key).
_keyed_dispatch_registries = DefaultDict[str, Dict[str, Tuple[Callable, Dict[Any, Callable]]]](dict)

def key_dispatched(key_fn:Optional[Callable[[Any], Any]]=None, *, key:Any=Default._) -> Callable[[Callable], Callable]:
  '''
  Decorator to register a method as the dispatched method for the specified key.
  Usage:
    @key_dispatched
  '''

  if key_fn is not None: # register new dispatcher name with default implementation.
    if not callable(key_fn):
      raise TypeError(f'argument to key_dispatched() must be key function; received {key_fn!r}')
    if key is not Default._:
      raise ValueError(f'@key_dispatch() takes either positional `key_fn` (decorating default method) or keyword `key`')
    if key_fn.__closure__ is not None: # type: ignore # Key function cannot be a closure.
      raise ValueError(f'@key_dispatch() `key_fn` cannot be a closure; recieved {key_fn!r}')

    def dflt_decorator(dflt_method:Callable)->Callable:
      _check_key_dispatched_method(dflt_method)
      # Check for collision.
      name = dflt_method.__name__
      module_registry = _keyed_dispatch_registries[dflt_method.__module__]
      if name in module_registry:
        raise DispatchKeyError(f'`keyed_dispatch` collision on method name: {name}')
      # Register.
      method_registry: Dict[Any, Callable] = {}
      # Create and return dispatcher method.
      @wraps(dflt_method)
      def dispatch(self, arg, *args, **kwargs):
        f = method_registry.get(key_fn(arg), dflt_method) # type: ignore
        return f(self, arg, *args, **kwargs)
      module_registry[name] = (dispatch, method_registry)
      return dispatch
    return dflt_decorator

  else: # key_fn is None; register implementation for key.
    if key is Default._:
      raise ValueError(f'@key_dispatch() takes either positional `key_fn` (decorating default method) or keyword `key`')

    def decorator(method:Callable) -> Callable:
      _check_key_dispatched_method(method)
      # Check for collision.
      name = method.__name__
      module_registry = _keyed_dispatch_registries[method.__module__]
      try: dispatch, method_registry = module_registry[name]
      except KeyError as e:
        raise DispatchKeyError(f'`keyed_dispatch` method name was not previously registered: {name}') from e
      if key in method_registry:
        raise DispatchKeyError(f'`keyed_dispatch` collision on key: {key}') # TODO: show previous implementation.
      method_registry[key] = method
      return dispatch
    return decorator


def _check_key_dispatched_method(method:Callable) -> None:
  if method.__closure__ is not None: # type: ignore # Method cannot be a closure because it is registered globally.
    raise ValueError(f'@key_dispatch() decorated method cannot be a closure; recieved {method!r}')
  sig = inspect.signature(method)
  pars = list(sig.parameters.values())
  if len(pars) < 2 or pars[0].name != 'self':
    raise TypeError(f'`keyed_dispatch` decorated method requires at least `self` and one additional parameter; received {method!r}')



#!/usr/bin/env python3
# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

# total hours in a timesheet.

# format is:

# (YYYY-)?MM-DD
# 12:00 begin some task.
# 18:30 change tasks.
# 24:00 end the task = 12:00.
# +$100 cost of materials.
# -$100 a payment.

import re
from argparse import ArgumentParser
from dataclasses import dataclass
from typing import List, Match, Optional

from ..io import outL, outZ


@dataclass
class Day:
  day: str
  minutes: int = 0

  def __str__(self) -> str:
    return '{}: {:>2}:{:02}'.format(self.day, *divmod(self.minutes, 60))


def main() -> None:
  parser = ArgumentParser(description='Validate timesheets.')
  parser.add_argument('timesheet', nargs='?', default='timesheet.txt')
  parser.add_argument('-rate', type=int, default=0)
  args = parser.parse_args()

  path = args.timesheet
  hourly_rate = args.rate

  days:List[Day] = []
  start_minutes:Optional[int] = None
  end_minutes:Optional[int]   = None
  total_minutes = 0
  total_payment = 0.0
  total_expense = 0.0

  valid = True

  try: f = open(path)
  except FileNotFoundError: exit(f'bad path: {path}')

  for line in f:
    l = line.rstrip('\n')
    outZ(f'{l:64}')

    day_match = day_re.match(line)
    if day_match:
      outL('(day)')
      if start_minutes is not None or end_minutes is not None:
        exit(f'timesheet error: previous day is missing end time.')
      days.append(Day(day_match[0]))
      continue

    time_match = time_re.match(line)
    if time_match:
      m = minutes_for(time_match)
      if start_minutes is None: start_minutes = m
      else: end_minutes = m # Cumulative from last start.
      outZ(f'|{m:4} ')

    subtotal_match = subtotal_re.search(line)
    if subtotal_match:
      if not time_match:
        outL()
        exit(f'timesheet error: subtotal line does not specify a time.')
      if start_minutes is None or end_minutes is None:
        outL()
        exit(f'timesheet error: subtotal line has invalid time: {subtotal_match[0]!r}')
      sub_minutes = end_minutes - start_minutes
      m = minutes_for(subtotal_match)
      outZ(f'= {sub_minutes:4}m')
      if m != sub_minutes:
        outZ(f' *** found: {m}; calculated: {sub_minutes}')
        valid = False
      if sub_minutes <= 0:
        outL()
        exit(f'timesheet error: subtototal is negative')
      days[-1].minutes += sub_minutes
      total_minutes += sub_minutes
      start_minutes = None
      end_minutes = None

    money_match = money_re.match(line)
    if money_match:
      s = ''.join(money_match.groups())
      i = float(s)
      if (i < 0):
        total_payment += i
      else:
        total_expense += i
      outZ(f'               {i: 10,.2f}')

    outL()


  hours, minutes = divmod(int(total_minutes), 60)
  time_expense = hourly_rate * total_minutes / 60
  total = time_expense + total_payment + total_expense
  if hourly_rate:
    hourly_string = ' @ {:0.2f}/hr = ${:,.2f}'.format(hourly_rate, time_expense)
  else:
    hourly_string = ''

  outL()
  outL(f'DAYS:')
  for day in days: outL(day)

  outL()
  outL(f'TOTAL HOURS:   {hours:2}:{minutes:02}{hourly_string}')
  outL(f'TOTAL EXPENSE: ${total_expense:,.2f}')
  outL(f'TOTAL PAYMENT: ${total_payment:,.2f}')
  outL(f'TOTAL:         ${total:,.2f}')

  if not valid:
    exit('*** INVALID ***')


def minutes_for(match:Match) -> int:
  return int(match.group(1)) * 60 + int(match.group(2))

day_re      = re.compile(r'(?:(\d\d\d\d)-)?(\d\d)-(\d\d)')
time_re     = re.compile(r'(\d\d):(\d\d) ')
subtotal_re = re.compile(r'= (\d{1,2}):(\d\d)')
money_re    = re.compile(r'([+-])\s*\$(\d+)(\.?\d*)')


if __name__ == '__main__': main()



#!/usr/bin/env python3
# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import re
from argparse import ArgumentParser
from dataclasses import dataclass, field
from shlex import quote as sh_quote
from typing import Any, Dict, List, Pattern, Set
from urllib.parse import urldefrag, urljoin

import html5_parser

from pithy.fs import (file_status, is_dir, is_file, make_dir, make_dirs, make_link, move_file, path_descendants, path_dir,
  path_ext, path_join, remove_path, remove_path_if_exists)
from pithy.io import errL, errSL, outL
from pithy.iterable import prefix_tree
from pithy.path_encode import COMP, OMIT, SPLIT, SQUASH, path_for_url
from pithy.string import clip_first_prefix
from pithy.task import runO


def main() -> None:
  arg_parser = ArgumentParser(description='Crawl URLs, using a regex filter for discovered links.')
  arg_parser.add_argument('seeds', nargs='+', help='URLs to crawl.')
  arg_parser.add_argument('-patterns', nargs='*', help='regex predicates that filter discovered links. '
    'A URL matching any pattern will be crawled. Defaults to the seed URLs.')
  arg_parser.add_argument('-clip-paths', nargs='*', default=[], help='path prefixes to clip each result with.')
  arg_parser.add_argument('-out', required=True, help='output directory.')
  arg_parser.add_argument('-force', action='store_true', help='always fetch URLs, even when cached results are present.')
  args = arg_parser.parse_args()

  dir = args.out
  seed0 = args.seeds[0]
  seeds = {clean_url(seed0, url) for url in args.seeds}

  ds = file_status(dir, follow=True)
  if ds is None:
    make_dirs(dir)
  elif not ds.is_dir:
    exit(f'output path exists and is not a directory: {dir!r}')

  if not args.patterns: # Default to seed URLs.
    args.patterns = [re.escape(s) for s in seeds]

  patterns:List[Pattern] = []
  for s in args.patterns:
    try: p = re.compile(s)
    except Exception as e: exit(f'bad pattern regex: {e}\n {s}')
    patterns.append(p)


  crawler = Crawler(
    dir=dir,
    patterns=patterns,
    clip_paths=args.clip_paths,
    remaining=set(seeds),
    visited=set(),
    force=args.force)

  crawler.crawl()
  describe_skipped(crawler.skipped)


@dataclass(frozen=True)
class Crawler:
  dir:str
  patterns:List[Pattern]
  clip_paths:List[str]
  remaining:Set[str]
  visited:Set[str]
  force:bool
  skipped:Set[str] = field(default_factory=set)
  max_redirects = 8

  def crawl(self) -> None:
    while self.remaining:
      url = self.remaining.pop()
      self.crawl_url(url=url)


  def crawl_url(self, url:str, redirects=0) -> None:
    assert url not in self.remaining
    assert url not in self.visited

    if redirects > self.max_redirects:
      errSL('too many redirects:', url)
      return

    self.visited.add(url)

    path = self.path_for_url(url)
    tmp_path = self.tmp_path_for_url(url)
    if is_dir(tmp_path, follow=False): exit(f'error: please remove the directory existing at path: {path}')
    if is_dir(tmp_path, follow=False): exit(f'error: please remove the directory existing at tmp path: {tmp_path}')

    if self.force or not is_file(path, follow=True):
      make_dirs(path_dir(tmp_path))
      remove_path_if_exists(tmp_path)

      cmd = ['curl', url, '-o', tmp_path]
      errSL('\ncrawl:', *[sh_quote (word) for word in cmd])
      cmd.extend(['--progress-bar', '--write-out', curl_output_fmt])
      output = runO(cmd)
      errL(output)
      results = parse_curl_output(output)
      code = results['http_code']

      if code in http_redirect_codes:
        redirect_url = clean_url(base=url, url=results['redirect_url'])
        if redirect_url == url:
          errSL(f'CIRCULAR redirect: {url} -> {redirect_url}')
          return
        errSL(f'redirect: {url} -> {redirect_url}')
        redirect_path = self.path_for_url(redirect_url)
        make_link(orig=redirect_path, link=path, allow_nonexistent=True, overwrite=True)
        remove_path(tmp_path)
        if redirect_url not in self.visited and redirect_url not in self.remaining:
          self.crawl_url(redirect_url, redirects=redirects+1)

      elif code in http_success_codes:
        move_file(tmp_path, to=path, overwrite=True)

      else:
        errSL('bad result:', code)
        return

      errSL(f'{url} : {path}')

    # Determine if we should scrape this document for links.
    url = url.partition('?')[0]
    ext = path_ext(url)
    should_scrape = (ext in ('', '.html', '.htm'))
    if should_scrape:
      self.try_scrape(url, path)


  def try_scrape(self, url:str, path:str) -> None:
    try: text = open(path).read()
    except IsADirectoryError: raise
    except Exception as e:
      errSL(f'{path}: could not read contents as text: {e}')
      return
    if html5_re.match(text):
      self.scrape_html5(url=url, text=text)
    else:
      raise Exception('non-html5 parsing is not supported.')


  def scrape_html5(self, url:str, text:str) -> None:
    html = html5_parser.parse(text, return_root=True, line_number_attr='#', sanitize_names=False)
    self.walk_html_hrefs(url=url, node=html)

  def walk_html_hrefs(self, url:str, node:Any) -> None:
    href = node.get('href')
    if href is not None:
      self.add_url(base=url, url=href)
    for child in node:
      self.walk_html_hrefs(url=url, node=child)


  def add_url(self, base:str, url:str) -> None:
    url = clean_url(base, url)
    if url in self.skipped or url in self.remaining or url in self.visited: return
    for p in self.patterns:
      if p.match(url):
        self.remaining.add(url)
        return
    #if url not in self.skipped: outSL('skipping:', url)
    self.skipped.add(url)


  def path_for_url(self, url:str) -> str:
    p = path_for_url(url, normalize=True, scheme=OMIT, host=COMP, path=SPLIT, query=SQUASH, fragment=OMIT)
    p = clip_first_prefix(p, self.clip_paths, req=False)
    return self.resolve_dir_collisions(path_join(self.dir, p))


  def tmp_path_for_url(self, url:str) -> str:
    p = path_for_url(url, normalize=True, scheme=OMIT, host=COMP, path=SQUASH, query=SQUASH, fragment=OMIT)
    return path_join(self.dir, p) + '.tmp'


  def resolve_dir_collisions(self, path:str) -> str:
    '''
    If any intermediate paths are occupied by files, rename those files and create directories.
    Return the path to use.
    '''
    # Resolve intermediates.
    for intermediate in path_descendants(self.dir, path, include_start=False, include_end=False):
      s = file_status(intermediate, follow=True)
      if s is None:
        make_dir(intermediate)
      elif not s.is_dir:
        tmp_path = path_join(self.dir, '_resolving_dir_collision.tmp')
        final_path = intermediate + '/+'
        move_file(intermediate, tmp_path)
        make_dir(intermediate)
        move_file(tmp_path, final_path)
    # Resolve tip.
    # TODO: improve confidence that this does the right thing in the presence of symlinks.
    s = file_status(path, follow=False)
    if s is None or not s.is_dir: return path
    path += '/+'
    if is_dir(path, follow=False):
      exit(f'error: path collision could not be resolved. Please remove the directory at {path!r}')
    return path


curl_output_fmt = '|'.join(f'{k}:%{{{k}}}' for k in [
  'http_code',
  'num_redirects',
  'content_type',
  'size_download',
  'time_total',
  'redirect_url',
])


def parse_curl_output(output:str) -> Dict[str,str]:
  triples = [word.partition(':') for word in output.split('|')]
  return { k:v for k,_,v in triples }


http_redirect_codes = {
  '301',
  '302',
  '303',
  '307',
  '308',
}

http_success_codes = {
  '200', # Standard.
  '203', # Transforming proxy is returning a modified version.
}


def clean_url(base:str, url:str='') -> str:
  res = urljoin(base, url)
  return urldefrag(res)[0] # type: ignore


def describe_skipped(skipped:Set[str]) -> None:
  'Describe URLs that were skipped by building a prefix tree.'
  tree = prefix_tree(_split_skipped(url) for url in skipped)
  _simplify(tree)
  outL('\nSkipped URLs:')
  _describe_skipped(tree, indent='')


def _split_skipped(url:str) -> List[str]:
  '''
  Split a skipped URL into parts for the purpose of building an informative prefix tree.
  We only omit the http and https schemes, because other schemes are unusual and should be reported.
  '''
  url = clip_first_prefix(url, ['https://', 'http://'], req=False)
  return url_parts_re.split(url)


def _simplify(tree:Dict) -> None:
  try: del tree[None]
  except KeyError: pass
  for child in tree.values(): _simplify(child)


def _describe_skipped(tree:Dict, indent:str) -> None:
  if len(tree) == 1:
    k, v = next(iter(tree.items()))
    _describe_skipped(v, indent=f'{indent}{k}/')
  elif tree and not any(tree.values()):
    outL(indent, ', '.join(k or repr(k) for k in tree))
  else:
    for k, v in sorted(tree.items()):
      outL(indent, k or repr(k))
      _describe_skipped(v, indent+'  ')


url_parts_re = re.compile(r'(?x) [/]+')

html5_re = re.compile(r'(?ix) \s* <!doctype \s+ html>')


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Base Markup type for Html, Svg, Xml, as well as legacy SGML formats.
'''

import re
from itertools import chain
from typing import (Any, Callable, Dict, FrozenSet, Generator, Iterable, Iterator, List, Match, Optional, Tuple, Type, TypeVar,
  Union, cast, overload)
from xml.etree.ElementTree import Element

from .desc import repr_lim
from .exceptions import ConflictingValues, DeleteNode, FlattenNode, MultipleMatchesError, NoMatchError
from .iterable import window_iter, window_pairs


# Handle lxml comments if available; these are produced by html5_parser.
try: from lxml.etree import Comment  # type: ignore
except ImportError: Comment = object() # Comment is a cyfunction (weirdly), so we can fallback to a dummy object.


_T = TypeVar('_T')

# Attr values are currently Any so that we can preserve exact numerical values.
MuAttrs = Dict[str,Any]
MuAttrItem = Tuple[str,Any]

MuChild = Union[str,'EscapedStr','Mu']
MuChildren = List[MuChild]
MuChildOrChildren = Union[MuChild,Iterable[MuChild]]

_Mu = TypeVar('_Mu', bound='Mu')
_MuChild = TypeVar('_MuChild', bound='MuChild')

MuPred = Callable[[_Mu],bool]
MuVisitor = Callable[[_Mu],None]
MuIterVisitor = Callable[[_Mu],Iterator[_T]]


class EscapedStr:
  def __init__(self, string:str) -> None:
    self.string = string


class Mu:
  '''
  Mu root class for building Html/Sgml/Xml document trees.
  Unlike xml.etree.ElementTree.Element, child nodes and text are interleaved.
  '''

  tag = '' # Subclasses can override the class tag, or give each instance its own tag attribute.
  tag_types:Dict[str,Type['Mu']] = {}
  inline_tags:FrozenSet[str] = frozenset()
  void_tags:FrozenSet[str] = frozenset()
  ws_sensitive_tags:FrozenSet[str] = frozenset()
  replaced_attrs:Dict[str,str] = {}

  __slots__ = ('attrs', 'ch', '_orig', '_parent')


  def __init__(self:_Mu, *, tag:str='', attrs:MuAttrs=None, ch:MuChildOrChildren=(), cl:Iterable[str]=None,
   _orig:_Mu=None, _parent:'Mu'=None, **kw_attrs:Any) -> None:
    '''
    Note: the initializer uses `attrs` dict and `ch` list references if provided, resulting in data sharing.
    This is done for two reasons:
    * avoid excess copying during deserialization from json, msgpack, or similar;
    * allow for creation of subtree nodes (with _orig/_parent set) that alias the `attr` and `ch` collections.

    Normally, nodes do not hold a reference to parent; this makes Mu trees acyclic.
    However, various Mu methods have a `traversable` option, which will return subtrees with the _orig/_parent refs set.
    Such "subtree nodes" can use the `next` and `prev` methods in addition to `pick` and friends.
    '''
    if attrs is None: attrs = {} # Important: use existing dict ref if provided.
    for k, v in kw_attrs.items():
      attrs[k.replace('_', '-')] = v
    self.attrs = attrs

    if cl is not None:
      if not isinstance(cl, str): cl = ' '.join(cl)
      if cl != attrs.setdefault('class', cl):
        raise ConflictingValues((attrs['class'], cl))

    if isinstance(ch, (str,EscapedStr,Mu)):
      self.ch:MuChildren = [ch]
    elif isinstance(ch, list):
      self.ch = ch # Important: use an existing list ref if provided. This allows subnodes to alias original contents.
    else:
      self.ch = list(ch)

    self._orig = _orig
    self._parent = _parent


  def __repr__(self) -> str: return f'{type(self).__name__}{self}'


  def __str__(self) -> str:
    subnode = '' if self._orig is None else '$'
    words = ''.join(chain(
      (xml_attr_summary(k, v, text_limit=32, all_attrs=False) for k, v in self.attrs.items()),
      (xml_child_summary(c, text_limit=32) for c in self.ch)))
    return f'<{subnode}{self.tag}:{words}>'


  def __delitem__(self, key:str) -> Any: del self.attrs[key]

  def __getitem__(self, key:str) -> Any: return self.attrs[key]

  def __setitem__(self, key:str, val:Any) -> Any: self.attrs[key] = val

  def get(self, key:str, default=None) -> Any: return self.attrs.get(key, default)

  def __iter__(self) -> Iterator[MuChild]: return iter(self.ch)


  @classmethod
  def from_raw(Class:Type[_Mu], raw:Dict) -> _Mu:
    'Create a Mu object (or possibly a subclass instance chosen by tag) from a raw data dictionary.'
    tag = raw['tag']
    attrs = raw['attrs']
    raw_children = raw['ch']
    if not isinstance(tag, str): raise ValueError(tag)
    if not isinstance(attrs, dict): raise ValueError(attrs)
    for k, v in attrs.items():
      if not isinstance(k, str):
        raise ValueError(f'Mu attr key must be `str`; received: {k!r}')
    ch:MuChildren = []
    for c in raw_children:
      if isinstance(c, (str, Mu)): ch.append(c)
      elif isinstance(c, dict): ch.append(Class.from_raw(c))
      else: raise ValueError(f'Mu child must be `str`, `Mu`, or `dict`; received: {c!r}')
    TagClass = Class.tag_types.get(tag, Class)
    return cast(_Mu, TagClass(tag=tag, attrs=attrs, ch=ch))


  @classmethod
  def from_etree(Class:Type[_Mu], el:Element) -> _Mu:
    '''
    Create an Mu object (possibly subclass by tag) from a standard library Mu element tree.
    Note: this handles lxml comment objects specially, by turning them into nodes with a '!COMMENT' tag.
    '''
    tag = el.tag
    if tag == Comment: tag = '!COMMENT' # Weird that the tag is an object, but this is what html5_parser produces.
    # Collect children.
    attrs = el.attrib
    ch:MuChildren = []
    text = el.text
    if text: ch.append(text)
    for child in el:
      ch.append(Class.from_etree(child))
      text = child.tail
      if text: ch.append(text)
    TagClass = Class.tag_types.get(tag, Class)
    return cast(_Mu, TagClass(tag=tag, attrs=attrs, ch=ch))


  @property
  def orig(self:_Mu) -> _Mu:
    'If this node is a query subnode, return the original; otherwise raise ValueError.'
    if self._orig is None: raise ValueError(f'node is not a subnode: {self}')
    return self._orig


  @property
  def parent(self) -> 'Mu':
    if self._parent is None: raise ValueError(f'node is not a subnode: {self}')
    return self._parent


  def subnode(self:_Mu, parent:'Mu') -> _Mu:
    'Create a subnode for `self` referencing the provided `parent`.'
    if self._orig is not None: raise ValueError(f'node is already a subnode: {self}')
    return type(self)(tag=self.tag, attrs=self.attrs, ch=self.ch, _orig=self, _parent=parent)


  def child_items(self, ws=False, traversable=False) -> Iterator[Tuple[int,MuChild]]:
    'Yield (index, child) pairs. If `ws` is False, then children that are purely whitespace will be filtered out.'
    for i, c in enumerate(self.ch):
      if isinstance(c, Mu):
        yield (i, (c.subnode(self) if traversable else c))
        continue
      if isinstance(c, EscapedStr):
        c = c.string
      if not ws and html_ws_re.fullmatch(c): continue
      yield (i, c)


  def children(self, ws=False, traversable=False) -> Iterator[MuChild]:
    'Yield child nodes and text. If `ws` is False, then children that are purely whitespace will be filtered out.'
    for c in self.ch:
      if isinstance(c, Mu):
        yield c.subnode(self) if traversable else c
        continue
      if isinstance(c, EscapedStr):
        c = c.string
      if not ws and html_ws_re.fullmatch(c): continue
      yield c


  def child_nodes(self, traversable=False) -> Iterator['Mu']:
    'Yield child Mu nodes.'
    return ((c.subnode(self) if traversable else c) for c in self.ch if isinstance(c, Mu))


  @property
  def has_substantial_children(self) -> bool:
    for c in self.ch:
      if isinstance(c, Mu): return True
      if isinstance(c, EscapedStr): c = c.string
      if c and not html_ws_re.fullmatch(c): return True
    return False


  @property
  def texts(self) -> Iterator[str]:
    for c in self.ch:
      if isinstance(c, Mu): yield from c.texts
      elif isinstance(c, EscapedStr): yield c.string
      else: yield c

  @property
  def text(self) -> str:
    return ''.join(self.texts)


  @property
  def cl(self) -> str: return str(self.attrs.get('class', ''))

  @cl.deleter
  def cl(self) -> None: del self.attrs['class']

  @cl.setter
  def cl(self, val:str) -> None: self.attrs['class'] = val


  @property
  def classes(self) -> List[str]: return cast(str, self.attrs.get('class', '')).split()

  @classes.deleter
  def classes(self) -> None: del self.attrs['class']

  @classes.setter
  def classes(self, val:Union[str, Iterable[str]]) -> None:
    if not isinstance(val, str): val = ' '.join(val)
    self.attrs['class'] = val


  def prepend_class(self, cl:str) -> None:
    try: existing = self.attrs['class']
    except KeyError: self.attrs['class'] = cl
    else: self.attrs['class'] = f'{cl} {existing}'


  def append_class(self, cl:str) -> None:
    try: existing = self.attrs['class']
    except KeyError: self.attrs['class'] = cl
    else: self.attrs['class'] = f'{existing} {cl}'


  @property
  def id(self) -> str: return str(self.attrs.get('id', ''))

  @id.setter
  def id(self, val:str) -> None: self.attrs['id'] = val

  @id.deleter
  def id(self) -> None: del self.attrs['id']


  def append(self, child:_MuChild) -> _MuChild:
    if isinstance(child, Mu) and child._orig is not None: child = child._orig
    if not isinstance(child, (str, Mu)): raise TypeError(child)
    self.ch.append(child)
    return child # type: ignore # The type of child._orig the same as child.


  def extend(self, children:Iterable[_MuChild]) -> None:
    for el in children: self.append(el)


  def _single(self, Child_type:Type[_Mu]) -> _Mu:
    for c in self.ch:
      if isinstance(c, Child_type): return c
    return self.append(Child_type())


  def clean(self, deep=True) -> None:
    # Consolidate consecutive strings.
    ch:List[MuChild] = []
    for c in self.ch:
      if isinstance(c, Mu):
        if deep: c.clean(deep)
      elif isinstance(c, str):
        if not c: continue # Do not append.
        if ch and isinstance(ch[-1], str): # Consolidate.
          ch[-1] += c
          continue # Do not append.
      else:
        raise ValueError(c) # Not (str, Mu).
      ch.append(c)

    inline_tags = self.inline_tags

    if self.tag not in self.ws_sensitive_tags:
      # Strip strings adjacent to block elements.
      for i, (p, c, n) in enumerate(window_iter(ch, width=3), 1):
        if not isinstance(c, str): continue
        assert isinstance(p, Mu)
        assert isinstance(n, Mu)
        if p.tag not in inline_tags: c = c.lstrip()
        if n.tag not in inline_tags: c = c.rstrip()
        ch[i] = c

      # If this element is a block, strip text at beginning and end.
      if ch and self.tag not in inline_tags:
        c0 = ch[0]
        if isinstance(c0, str): ch[0] = c0.lstrip()
        cl = ch[-1]
        if isinstance(cl, str): ch[-1] = cl.rstrip()

      ch = [c for c in ch if c] # Filter now-empty text elements.

      # Reduce remaining, repeated whitespace down to single '\n' and ' ' characters.
      # https://www.w3.org/TR/CSS22/text.html#white-space-model
      # https://drafts.csswg.org/css-text-3/#white-space-phase-1
      for i in range(len(ch)):
        c = ch[i]
        if isinstance(c, str):
          ch[i] = html_ws_re.sub(html_ws_replacement, c)

    self.ch[:] = ch # Mutate the original array beacuse it may be aliased by subnodes.



  # Picking and finding.

  @overload
  def pick_all(self, type_or_tag:Type[_Mu], *, cl:str='', text:str='', traversable=False, **attrs:str) -> Iterator[_Mu]: ...

  @overload
  def pick_all(self, type_or_tag:str='', *, cl:str='', text:str='', traversable=False, **attrs:str) -> Iterator['Mu']: ...

  def pick_all(self, type_or_tag='', *, cl:str='', text:str='', traversable=False, **attrs:str):
    pred = xml_pred(type_or_tag=type_or_tag, cl=cl, text=text, attrs=attrs)
    return ((c.subnode(self) if traversable else c) for c in self.ch if isinstance(c, Mu) and pred(c))


  @overload
  def find_all(self, type_or_tag:Type[_Mu], *, cl:str='', text:str='', traversable=False, **attrs:str) -> Iterator[_Mu]: ...

  @overload
  def find_all(self, type_or_tag:str='', *, cl:str='', text:str='', traversable=False, **attrs:str) -> Iterator['Mu']: ...

  def find_all(self, type_or_tag='', *, cl:str='', text:str='', traversable=False, **attrs:str):
    pred = xml_pred(type_or_tag=type_or_tag, cl=cl, text=text, attrs=attrs)
    if text: return self._find_all_text(pred, traversable)
    else: return self._find_all(pred, traversable)

  def _find_all(self, pred:MuPred, traversable:bool) -> Iterator['Mu']:
    for c in self.ch:
      if isinstance(c, Mu):
        if pred(c): yield (c.subnode(self) if traversable else c)
        yield from c._find_all(pred, traversable) # Always search ch. TODO: use generator send() to let consumer decide?

  def _find_all_text(self, pred:MuPred, traversable:bool) -> Generator['Mu',None,bool]:
    '''
    Use post-order algorithm to find matching text, and do not search parents of matching children.
    This is desirable because the calculation of text is expensive
    and the caller most likely does not want nodes that contain each other.
    '''
    found_match = False
    for c in self.ch:
      if isinstance(c, Mu):
        child_match = yield from c._find_all_text(pred, traversable)
        if child_match:
          found_match = True
        elif pred(c):
          found_match = True
          yield (c.subnode(self) if traversable else c)
    return found_match


  @overload
  def pick_first(self, type_or_tag:Type[_Mu], *, cl:str='', text:str='', traversable=False, **attrs:str) -> _Mu: ...

  @overload
  def pick_first(self, type_or_tag:str='', *, cl:str='', text:str='', traversable=False, **attrs:str) -> 'Mu': ...

  def pick_first(self, type_or_tag='', *, cl:str='', text:str='', traversable=False, **attrs:str):
    pred = xml_pred(type_or_tag=type_or_tag, cl=cl, text=text, attrs=attrs)
    for c in self.ch:
      if isinstance(c, Mu) and pred(c): return (c.subnode(self) if traversable else c)
    raise NoMatchError(self, fmt_xml_predicate_args(type_or_tag, cl, text, attrs))


  @overload
  def find_first(self, type_or_tag:Type[_Mu], *, cl:str='', text:str='', traversable=False, **attrs:str) -> _Mu: ...

  @overload
  def find_first(self, type_or_tag:str='', *, cl:str='', text:str='', traversable=False, **attrs:str) -> 'Mu': ...

  def find_first(self, type_or_tag='', *, cl:str='', text:str='', traversable=False, **attrs:str):
    for c in self.find_all(type_or_tag=type_or_tag, cl=cl, text=text, traversable=traversable, **attrs):
      return c
    raise NoMatchError(self, fmt_xml_predicate_args(type_or_tag, cl, text, attrs))


  @overload
  def pick(self, type_or_tag:Type[_Mu], *, cl:str='', text:str='', traversable=False, **attrs:str) -> _Mu: ...

  @overload
  def pick(self, type_or_tag:str='', *, cl:str='', text:str='', traversable=False, **attrs:str) -> 'Mu': ...

  def pick(self, type_or_tag='', *, cl:str='', text:str='', traversable=False, **attrs:str):
    first_match:Optional[Mu] = None
    for c in self.pick_all(type_or_tag=type_or_tag, cl=cl, text=text, traversable=traversable, **attrs):
      if first_match is None: first_match = c
      else:
        args_msg = fmt_xml_predicate_args(type_or_tag, cl, text, attrs)
        subsequent_match = c # Alias improves readablity of the following line in stack traces.
        raise MultipleMatchesError(self, args_msg, first_match, subsequent_match)
    if first_match is None:
      raise NoMatchError(self, fmt_xml_predicate_args(type_or_tag, cl, text, attrs))
    return first_match


  @overload
  def find(self, type_or_tag:Type[_Mu], *, cl:str='', text:str='', traversable=False, **attrs:str) -> _Mu: ...

  @overload
  def find(self, type_or_tag:str='', *, cl:str='', text:str='', traversable=False, **attrs:str) -> 'Mu': ...

  def find(self, type_or_tag='', *, cl:str='', text:str='', traversable=False, **attrs:str):
    first_match:Optional[Mu] = None
    for c in self.find_all(type_or_tag=type_or_tag, cl=cl, text=text, traversable=traversable, **attrs):
      if first_match is None: first_match = c
      else:
        args_msg = fmt_xml_predicate_args(type_or_tag, cl, text, attrs)
        subsequent_match = c # Alias improves readablity of the following line in stack traces.
        raise MultipleMatchesError(self, args_msg, first_match, subsequent_match)
    if first_match is None:
      raise NoMatchError(self, fmt_xml_predicate_args(type_or_tag, cl, text, attrs))
    return first_match


  # Traversal.

  @overload
  def next(self, type_or_tag:Type[_Mu], *, cl:str='', text:str='', traversable=False, **attrs:str) -> _Mu: ...

  @overload
  def next(self, type_or_tag:str='', *, cl:str='', text:str='', traversable=False, **attrs:str) -> 'Mu': ...

  def next(self, type_or_tag='', *, cl:str='', text:str='', traversable=False, **attrs:str):
    if self._orig is None or self._parent is None: raise ValueError(f'cannot traverse non-subnode: {self}')
    pred = xml_pred(type_or_tag=type_or_tag, cl=cl, text=text, attrs=attrs)
    found_orig = False
    for c in self._parent.ch:
      if not isinstance(c, Mu): continue
      if found_orig:
        if pred(c): return (c.subnode(self._parent) if traversable else c)
      elif c is self._orig:
        found_orig = True
    if not found_orig: raise ValueError('node was removed from parent')
    raise NoMatchError(self, fmt_xml_predicate_args(type_or_tag, cl, text, attrs))


  @overload
  def prev(self, type_or_tag:Type[_Mu], *, cl:str='', text:str='', traversable=False, **attrs:str) -> _Mu: ...

  @overload
  def prev(self, type_or_tag:str='', *, cl:str='', text:str='', traversable=False, **attrs:str) -> 'Mu': ...

  def prev(self, type_or_tag='', *, cl:str='', text:str='', traversable=False, **attrs:str):
    if self._orig is None or self._parent is None: raise ValueError(f'cannot traverse non-subnode: {self}')
    pred = xml_pred(type_or_tag=type_or_tag, cl=cl, text=text, attrs=attrs)
    found_orig = False
    for c in reversed(self._parent.ch):
      if not isinstance(c, Mu): continue
      if found_orig:
        if pred(c): return (c.subnode(self._parent) if traversable else c)
      elif c is self._orig:
        found_orig = True
    if not found_orig: raise ValueError('node was removed from parent')
    raise NoMatchError(self, fmt_xml_predicate_args(type_or_tag, cl, text, attrs))


  # Text.

  def summary_texts(self, _needs_space:bool=True) -> Generator[str,None,bool]:
    for child in self.ch:
      if isinstance(child, Mu):
        _needs_space = yield from child.summary_texts(_needs_space=_needs_space)
        continue
      for m in html_ws_split_re.finditer(str(child)):
        if m.lastgroup == 'space':
          if _needs_space:
            yield ' '
            _needs_space = False
        else:
          yield m[0]
          _needs_space = True
    return _needs_space


  def summary_text(self, limit=0) -> str:
    if not limit: return ''.join(self.summary_texts())
    parts:List[str] = []
    length = 0
    for part in self.summary_texts():
      parts.append(part)
      length += len(part)
      if length > limit: break
    return ''.join(parts)[:limit]


  # Text summary.

  def summarize(self, levels=1, indent=0, all_attrs=True) -> str:
    nl_indent = '\n' + '  ' * indent
    return ''.join(self._summarize(levels, nl_indent, all_attrs=all_attrs))

  def _summarize(self, levels:int, nl_indent:str, all_attrs:bool) -> Iterator[str]:
    if levels <= 0:
      yield str(self)
    else:
      subnode = '' if self._orig is None else '$'
      attr_words = ''.join(xml_attr_summary(k, v, text_limit=32, all_attrs=all_attrs) for k, v in self.attrs.items())
      nl_indent1 = nl_indent + '  '
      yield f'<{subnode}{self.tag}:{attr_words}'
      for c in self.ch:
        yield nl_indent1
        if isinstance(c, Mu):
          yield from c._summarize(levels-1, nl_indent1, all_attrs)
        else:
          yield repr(c)
      yield '>'


  def discard(self, attr:str) -> None:
    try: del self.attrs[attr]
    except KeyError: pass


  def visit(self, *, pre:MuVisitor=None, post:MuVisitor=None, traversable=False) -> None:
    if pre is not None: pre(self)

    modified_children:List[MuChild] = []
    first_mod_idx:Optional[int] = None
    for i, c in enumerate(self.ch):
      if isinstance(c, Mu):
        if traversable: c = c.subnode(self)
        try: c.visit(pre=pre, post=post, traversable=traversable)
        except DeleteNode:
          if first_mod_idx is None: first_mod_idx = i
          continue
        except FlattenNode:
          if first_mod_idx is None: first_mod_idx = i
          modified_children.extend(c.ch) # Insert children in place of `c`.
          continue
      if first_mod_idx is not None:
        modified_children.append(c)
    if first_mod_idx is not None:
      self.ch[first_mod_idx:] = modified_children

    if post is not None: post(self)


  def iter_visit(self, *, pre:MuIterVisitor=None, post:MuIterVisitor=None, traversable=False) -> Iterator[_T]:
    if pre is not None: yield from pre(self)

    for i, c in enumerate(self.ch):
      if isinstance(c, Mu):
        if traversable: c = c.subnode(self)
        yield from c.iter_visit(pre=pre, post=post, traversable=traversable)

    if post is not None: yield from post(self)


  # Rendering.


  def esc_attr_val(self, val:str) -> str: raise NotImplementedError

  def esc_text(self, text:str) -> str: raise NotImplementedError

  def fmt_attr_items(self, items:Iterable[Tuple[str,Any]]) -> str:
    'Return a string that is either empty or with a leading space, containing all of the formatted items.'
    parts: List[str] = []
    for k, v in items:
      k = self.replaced_attrs.get(k, k)
      if v in (None, True, False): v = str(v).lower()
      parts.append(f' {k}="{self.esc_attr_val(str(v))}"')
    return ''.join(parts)


  def render(self, newline=True) -> Iterator[str]:
    'Render the tree as a stream of text lines.'
    yield from self._render()
    if newline: yield '\n'


  def _render(self) -> Iterator[str]:
    'Recursive helper to `render`.'

    if self.void_tags:
      self_closing = self.tag in self.void_tags
      if self_closing and self.ch: raise ValueError(self)
    else:
      self_closing = not self.ch

    attrs_str = self.fmt_attr_items(self.attrs.items())
    head_slash = '/' if self_closing else ''
    yield f'<{self.tag}{attrs_str}{head_slash}>'
    if self_closing: return

    child_newlines = (
      len(self.ch) > 1 and
      (self.tag not in self.ws_sensitive_tags) and
      (self.tag not in self.inline_tags))

    def is_block(el:MuChild) -> bool: return isinstance(el, Mu) and (el.tag not in self.inline_tags)

    if child_newlines:
      yield '\n'
    for child, next_child in window_pairs(self.ch):
      if isinstance(child, Mu):
        yield from child._render()
      elif isinstance(child, EscapedStr):
        yield child.string
      else:
        yield self.esc_text(child)
      if child_newlines and (is_block(child) or next_child is None or is_block(next_child)):
        yield '\n'

    yield f'</{self.tag}>'


  def render_str(self, newline=True) -> str:
    'Render the tree into a single string.'
    return ''.join(self.render(newline=newline))



def xml_attr_summary(key:str, val:Any, *, text_limit:int, all_attrs:bool) -> str:
  ks = key if _word_re.fullmatch(key) else repr(key)
  if all_attrs or key in ('id', 'class'): return f' {ks}={repr_lim(val, text_limit)}' # Show id and class values.
  return f' {ks}=…' # Omit other attribute values.


def xml_child_summary(child:MuChild, text_limit:int) -> str:
  if isinstance(child, Mu):
    text = child.summary_text(limit=text_limit)
    if text: return f' {child.tag}:{repr_lim(text, limit=text_limit)}'
    return ' ' + child.tag
  if isinstance(child, EscapedStr):
    child = child.string
  text = html_ws_re.sub(newline_or_space_for_ws, child)
  return ' ' + repr_lim(text, limit=text_limit)


def xml_pred(type_or_tag:Union[str,Type[_Mu]]='', *, cl:str='', text:str='', attrs:Dict[str,Any]={}) -> MuPred:
  'Update _attrs with items from other arguments, then construct a predicate that tests Mu nodes.'

  tag_pred:Callable
  if not type_or_tag: tag_pred = lambda node: True
  elif isinstance(type_or_tag, type): tag_pred = lambda node: isinstance(node, type_or_tag) # type: ignore
  else: tag_pred = lambda node: node.tag == type_or_tag

  def predicate(node:Mu) -> bool:
    return (
      tag_pred(node) and
      (not cl or cl in node.classes) and
      all(node.attrs.get(k.replace('_', '-')) == v for k, v in attrs.items()) and
      (not text or text in node.text))

  return predicate


def fmt_xml_predicate_args(type_or_tag:Union[Type,str], cl:str, text:str, attrs:Dict[str,str]) -> str:
  words:List[str] = []
  if type_or_tag: words.append(f'`{type_or_tag.__name__}`' if isinstance(type_or_tag, type) else repr(type_or_tag))
  if cl: words.append(f'cl={cl!r}')
  for k, v in attrs.items(): words.append(xml_attr_summary(k, v, text_limit=0, all_attrs=True).lstrip())
  if text: words.append(f'…{text!r}…')
  return ' '.join(words)


def newline_or_space_for_ws(match:Match) -> str:
  return '\n' if '\n' in match[0] else ' '


def html_ws_replacement(m:Match) -> str:
  return '\n' if '\n' in m[0] else ' '


# HTML defines ASCII whitespace as "U+0009 TAB, U+000A LF, U+000C FF, U+000D CR, or U+0020 SPACE."
html_ws_re = re.compile(r'[\t\n\f\r ]+')
html_ws_split_re = re.compile(r'(?P<space>[\t\n\f\r ])|[^\t\n\f\r ]+')

_word_re = re.compile(r'[-\w]+')



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import sqlite3
from typing import Any, Dict, Iterable, Iterator, List, NamedTuple, Optional, Sequence, Tuple, Type, cast
from .json import render_json


class SqliteError(Exception): pass


class Cursor(sqlite3.Cursor):

  def execute(self, query: str, args:Iterable=()) -> 'Cursor':
    try: return cast('Cursor', super().execute(query, args))
    except sqlite3.Error as e:
      raise SqliteError(f'SQLite error; query: {query!r}') from e

  def run(self, *sql: str, **args: Any) -> 'Cursor':
    '''
    Execute a query, joining multiple pieces of `sql` into a single query string, with values provided by keyword arguments.
    Values whose types are not sqlite compatible are automatically converted to Json.
    '''
    query = ' '.join(sql)
    for k, v in args.items(): # Convert non-native values to Json.
      if not isinstance(v, py_to_sql_types_tuple):
        args[k] = render_json(v, indent=None)
    return self.execute(query, args)


  def insert(self, *, with_='', or_='FAIL', table: str, fields: Tuple[str, ...], sql: str, args: Any) -> None:
    '''
    Execute an insert statement synthesized from `table`, `fields`, and `sql`.
    '''
    assert or_ in {'ABORT', 'FAIL', 'IGNORE', 'REPLACE', 'ROLLBACK'}
    fields_str = ', '.join(fields)
    with_space = ' ' if with_ else ''
    sql = f'{with_}{with_space}INSERT OR {or_} INTO {table} ({fields_str}) {sql}'
    # TODO: cache sql with (with_, or_, table, fields, sql) key.
    self.execute(sql, args)


  def insert_seq(self, *, with_='', or_='FAIL', table: str, fields: Tuple[str, ...], args: Sequence[Any]) -> None:
    '''
    Execute an insert statement inserting the sequence `args`, synthesized from `table`, and `fields`.
    '''
    assert len(fields) == len(args)
    placeholders = ','.join('?' for _ in fields)
    self.insert(with_=with_, or_=or_, table=table, fields=fields, sql=f'VALUES ({placeholders})', args=args)


  def insert_dict(self, *, with_='', or_='FAIL', table:str, fields:Tuple[str, ...],
   args:Dict[str, Any], defaults:Dict[str, Any]={}, **overrides:Any) -> None:
    '''
    Execute an insert statement inserting the dictionary `args`, synthesized from `table`, `fields`.
    Values are pulled by name in order of preference from keyword argument overrides, the `args` dictionary, and `defaults`;
    a KeyError is raised if one of the fields is not provided in any of these sources.
    '''
    def arg_for(f: str) -> Any:
      try: return overrides[f]
      except KeyError: pass
      try: return args[f]
      except KeyError: pass
      return defaults[f]
    args_seq = [_default_to_json(arg_for(f)) for f in fields]
    self.insert_seq(with_=with_, or_=or_, table=table, fields=fields, args=args_seq)


  def select(self, *sql: str, **args: Any) -> 'Cursor':
    'Execute a SELECT query.'
    return self.run('SELECT', *sql, **args)

  def select_opt(self, *sql: str, **args: Any) -> Optional[List[Any]]:
    'Execute a SELECT query, returning a single row or None.'
    return self.run('SELECT', *sql, **args).fetchone() # type: ignore

  def select_col(self, *sql: str, **args: Any) -> Iterator[Any]:
    'Execute a SELECT query, returning column 0 of each result row.'
    for row in self.run('SELECT', *sql, **args):
      assert len(row) == 1
      yield row[0]

  def contains(self, table: str, *criteria: str, **args: Any) -> bool:
    'Execute a SELECT query, returning True if the `criteria` SQL result in a row.`'
    for row in self.run('SELECT 1 FROM', table, 'WHERE', *criteria, 'LIMIT 1', **args):
      return True
    return False


class Connection(sqlite3.Connection):

  def __init__(self, path: str, uri:bool=False) -> None:
    super().__init__(path, uri=uri) # type: ignore
    self.row_factory = sqlite3.Row # default for convenience.
    #self.stmt_cache = {}

  def cursor(self, factory:Optional[type]=None) -> Cursor:
    if factory is None:
      factory = Cursor
    assert issubclass(factory, Cursor)
    return cast(Cursor, super().cursor(factory))

  def run(self, *sql: str, **args: Any) -> Cursor:
    return self.cursor().run(*sql, **args)

  def insert(self, *, with_='', or_='FAIL', table: str, fields: Tuple[str, ...], sql: str, args: Any) -> None:
    return self.cursor().insert(with_=with_, or_=or_, table=table, fields=fields, sql=sql, args=args)

  def insert_seq(self, *, with_='', or_='FAIL', table: str, fields: Tuple[str, ...], args: Sequence[Any]) -> None:
    return self.cursor().insert_seq(with_=with_, or_=or_, table=table, fields=fields, args=args)

  def insert_dict(self, *, with_='', or_='FAIL', table:str, fields:Tuple[str, ...],
   args:Dict[str, Any], defaults:Dict[str, Any]={}, **overrides:Any) -> None:
    return self.cursor().insert_dict(with_=with_, or_=or_, table=table, fields=fields, args=args, defaults=defaults, **overrides)

  def select(self, *sql: str, **args: Any) -> Cursor:
    return self.select(*sql, **args)

  def select_opt(self, *sql: str, **args: Any) -> Optional[List[Any]]:
    return self.select_opt(*sql, **args)

  def select_col(self, *sql: str, **args: Any) -> Iterator[Any]:
    return self.select_col(*sql, **args)


  def contains(self, table: str, *criteria: str, **args: Any) -> bool:
    return self.contains(table, *criteria, **args)



def sql_col_names(dataclass: type) -> str:
  '''
  Given a dataclass or NamedTuple subclass, return a string of comma-separated field names.
  '''
  return ', '.join(fields_of(dataclass))


def sql_col_placeholders(dataclass: type) -> str:
  '''
  Given a dataclass or NamedTuple subclass, return a string of comma-separated SQL named placeholders.
  '''
  return ', '.join(f':{n}' for n in fields_of(dataclass))


def sql_col_decls(class_: Type[NamedTuple], primary: str) -> str:
  '''
  Given a dataclass or NamedTuple subclass, yield a sequence of SQL column declarations for use in a CREATE TABLE statement.
  '''
  decls = []
  for n, static_type in class_._field_types.items():
    # Currently supports primitive types and their optionals, and Json.
    try: sql_type = py_to_sql_types[static_type]
    except KeyError:
      try: unwrapped_type = _wrapped_type_for_optional(static_type)
      except TypeError: sql_type = 'TEXT'
      else: sql_type = py_to_sql_types.get(unwrapped_type, 'TEXT')
    suffix = f' PRIMARY KEY' if n == primary else ''
    decls.append(f'{n} {sql_type}{suffix}')
  return ', '.join(decls)


def _wrapped_type_for_optional(static_type: type) -> type:
  # Optionals are really unions, which are a pain to work with at runtime.
  try: meta_class_name = static_type.__class__.__name__
  except AttributeError as e: raise TypeError(static_type) from e
  if meta_class_name != '_Union': raise TypeError(static_type)
  members = static_type.__args__ # type: ignore
  if len(members) != 2 or NoneType not in members: raise TypeError(static_type) # type: ignore
  return [m for m in members if m is not NoneType][0] # type: ignore


def _default_to_json(obj: Any) -> Any:
  if isinstance(obj, py_to_sql_types_tuple): return obj
  return render_json(obj)


def fields_of(class_: type) -> Tuple[str, ...]:
  if issubclass(class_, NamedTuple): return class_._fields
  # TODO: support dataclasses.
  raise TypeError(class_)


NoneType = type(None)

py_to_sql_types: Dict[type, str] = {
  int : 'INTEGER',
  float: 'REAL',
  str : 'TEXT',
  bool : 'INT',
  bytes: 'BLOB',
  type(None): 'BLOB', # blob affinity has no conversion preference, so is most appropriate for unknown types.
}

py_to_sql_types_tuple = tuple(py_to_sql_types)

sql_to_py_types = { sql : py for (py, sql) in py_to_sql_types.items() }



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from os import DirEntry, stat as _stat, stat_result as StatResult
from stat import (S_IFBLK, S_IFCHR, S_IFDIR, S_IFIFO, S_IFLNK, S_IFMT, S_IFREG, S_IFSOCK, S_IRGRP, S_IROTH, S_IRUSR, S_IRWXG,
  S_IRWXO, S_IRWXU, S_ISGID, S_ISUID, S_ISVTX, S_IWGRP, S_IWOTH, S_IWUSR, S_IXGRP, S_IXOTH, S_IXUSR)
from typing import NamedTuple, Optional

from .path import Path, PathOrFd


class FileStatus(NamedTuple):
  '''
  Reinterpretation of standard stat_result:
  * nicer names for the fields;
  * omits atime and atime_ns (which are not useful due to modern `noatime` semantics);
  * splits `mode` into `type` and `perms` fields.
  '''
  birth_time: int
  block_count: int
  block_size: int
  ctime: float
  ctime_ns: int
  device: int
  device_type: int
  flags: int
  generation: int
  group_id: int
  inode: int
  link_count: int
  mode: int
  mtime: float
  mtime_ns: int
  path: str
  perms: int
  size: int
  type: int
  user_id: int

  @staticmethod
  def from_stat_result(path: str, stat_result: StatResult) -> 'FileStatus':
    s = stat_result
    mode = s.st_mode
    return FileStatus(
      birth_time=s.st_birthtime,
      block_count=s.st_blocks,
      block_size=s.st_blksize,
      ctime=s.st_ctime,
      ctime_ns=s.st_ctime_ns,
      device=s.st_dev,
      device_type=s.st_rdev,
      flags=s.st_flags,
      generation=s.st_gen,
      group_id=s.st_gid,
      inode=s.st_ino,
      link_count=s.st_nlink,
      mode=mode,
      mtime=s.st_mtime,
      mtime_ns=s.st_mtime_ns,
      path=path,
      perms=(PERM_MASK&mode),
      size=s.st_size,
      type=(TYPE_MASK&mode),
      user_id=s.st_uid)

  @classmethod
  def from_dir_entry(class_, entry: DirEntry, *, follow:bool) -> Optional['FileStatus']:
    return class_.from_stat_result(entry.path, entry.stat(follow_symlinks=follow))


  # File type tests derived from Python's stat.py.

  @property
  def is_block_device(self) -> bool: return self.type == S_IFBLK

  @property
  def is_character_device(self) -> bool: return self.type == S_IFCHR

  @property
  def is_dir(self) -> bool: return self.type == S_IFDIR

  @property
  def is_fifo(self) -> bool: return self.type == S_IFIFO

  @property
  def is_file(self) -> bool: return self.type == S_IFREG

  @property
  def is_link(self) -> bool: return self.type == S_IFLNK

  @property
  def is_socket(self) -> bool: return self.type == S_IFSOCK


  # Permissions.

  @property
  def is_suid(self) -> bool: return bool(self.perms & S_ISUID)

  @property
  def is_guid(self) -> bool: return bool(self.perms & S_ISGID)

  @property
  def is_sticky(self) -> bool: return bool(self.perms & S_ISVTX)

  @property
  def owner_perms(self) -> int: return self.perms & S_IRWXU

  @property
  def group_perms(self) -> int: return self.perms & S_IRWXG

  @property
  def other_perms(self) -> int: return self.perms & S_IRWXO

  @property
  def is_owner_readable(self) -> bool: return bool(self.perms & S_IRUSR)

  @property
  def is_owner_writeable(self) -> bool: return bool(self.perms & S_IWUSR)

  @property
  def is_owner_executable(self) -> bool: return bool(self.perms & S_IXUSR)

  @property
  def is_group_readable(self) -> bool: return bool(self.perms & S_IRGRP)

  @property
  def is_group_writeable(self) -> bool: return bool(self.perms & S_IWGRP)

  @property
  def is_group_executable(self) -> bool: return bool(self.perms & S_IXGRP)

  @property
  def is_other_readable(self) -> bool: return bool(self.perms & S_IROTH)

  @property
  def is_other_writeable(self) -> bool: return bool(self.perms & S_IWOTH)

  @property
  def is_other_executable(self) -> bool: return bool(self.perms & S_IXOTH)


  # Descriptions.

  @property
  def type_char(self) -> str: return _type_chars[self.type]

  @property
  def type_desc(self) -> str: return _type_descs[self.type]


  @property
  def perms_string(self) -> str:
    parts = [] # user, group, other.
    for part_char, part_table in _perm_chars:
        chars = part_char
        for bit, char in part_table:
          chars += char if (self.perms & bit) else '-'
        parts.append(chars)
    if self.is_suid: parts.append('suid')
    if self.is_guid: parts.append('sgid')
    if self.is_sticky: parts.append('sticky')
    return ' '.join(parts)

  @property
  def mode_str(self) -> str:
    return f'{self.type_char} {self.perms_string}'


def dir_entry_type_char(entry: DirEntry) -> str:
  '''
  Return a single uppercase letter string denoting the file type of the DirEntry.
  Because DirEntry does not expose the complete `stat` type field,
  this is limited to `D`, `F`, `L`, or `U` for other/unknown type,
  and thus is not necessarily equal to the letter obtained from `FileStatus.type_char`.
  '''
  if entry.is_symlink(): return 'L'
  if entry.is_dir(): return 'D'
  if entry.is_file(): return 'F'
  return 'U'


def file_ctime(path_or_fd:PathOrFd, *, follow:bool) -> float: return _stat(path_or_fd, follow_symlinks=follow).st_ctime

def file_inode(path_or_fd:PathOrFd, *, follow:bool) -> int: return _stat(path_or_fd, follow_symlinks=follow).st_ino

def file_permissions(path_or_fd:PathOrFd, *, follow:bool=True) -> int:
  # TODO: might be useful to OR the permissions of a symlink and its destination together; perhaps a separate function.
  return _stat(path_or_fd, follow_symlinks=follow).st_mode & PERM_MASK

def file_size(path_or_fd:PathOrFd, *, follow:bool=True) -> int: return _stat(path_or_fd, follow_symlinks=follow).st_size

def file_stat(path_or_fd:PathOrFd, *, follow:bool) -> StatResult: return _stat(path_or_fd, follow_symlinks=follow)

def file_status(path_or_fd:PathOrFd, *, follow:bool, raises=False) -> Optional[FileStatus]:
  try: s = _stat(path_or_fd, follow_symlinks=follow)
  except FileNotFoundError:
    if raises: raise
    return None
  path = '' if isinstance(path_or_fd, int) else str(path_or_fd)
  return FileStatus.from_stat_result(path, s)

def file_mtime(path_or_fd:PathOrFd, *, follow:bool) -> float:
  return _stat(path_or_fd, follow_symlinks=follow).st_mtime

def file_mtime_or_zero(path_or_fd:str, *, follow:bool) -> float:
  try: return file_mtime(path_or_fd, follow=follow)
  except FileNotFoundError: return 0

def is_dir(path_or_fd:Path, *, follow:bool, raises=False) -> Optional[bool]:
  try: s = _stat(path_or_fd, follow_symlinks=follow)
  except FileNotFoundError:
    if raises: raise
    return None
  return s.st_mode&TYPE_MASK == S_IFDIR

def is_file(path_or_fd:Path, *, follow:bool, raises=False) -> Optional[bool]:
  try: s = _stat(path_or_fd, follow_symlinks=follow)
  except FileNotFoundError:
    if raises: raise
    return None
  return s.st_mode&TYPE_MASK == S_IFREG

def is_file_executable_by_owner(path_or_fd:Path, *, follow:bool=True) -> bool:
  return bool(file_permissions(path_or_fd, follow=follow) & S_IXUSR)

def is_link(path_or_fd:Path, *, raises=False) -> Optional[bool]:
  try: s = _stat(path_or_fd, follow_symlinks=False)
  except FileNotFoundError:
    if raises: raise
    return None
  return s.st_mode&TYPE_MASK == S_IFLNK

def is_link_to_dir(path_or_fd:Path, *, raises=False) -> Optional[bool]:
  return is_link(path_or_fd, raises=raises) and is_dir(path_or_fd, follow=True, raises=raises)

def is_link_to_file(path_or_fd:Path, *, raises=False) -> Optional[bool]:
  return is_link(path_or_fd, raises=raises) and is_file(path_or_fd, follow=True, raises=raises)

def is_mount(path_or_fd:Path, *, follow:bool=True, raises=False) -> Optional[bool]:
  try: s = _stat(path_or_fd, follow_symlinks=follow)
  except FileNotFoundError:
    if raises: raise
    return None
  return s.st_mode&TYPE_MASK == S_IFMT

def is_sticky(path_or_fd:Path, *, follow:bool, raises=False) -> Optional[bool]:
  try: s = _stat(path_or_fd, follow_symlinks=follow)
  except FileNotFoundError:
    if raises: raise
    return None
  return bool(s.st_mode&S_ISVTX)

def path_exists(path_or_fd:Path, *, follow:bool) -> Optional[bool]:
  try: _stat(path_or_fd, follow_symlinks=follow)
  except FileNotFoundError: return False
  else: return True


PERM_MASK = 0o7777
TYPE_MASK = 0o170000


_type_chars = {
  S_IFBLK   : 'B',
  S_IFCHR   : 'C',
  S_IFDIR   : 'D',
  S_IFIFO   : 'P',
  S_IFLNK   : 'L',
  S_IFREG   : 'F', # traditionally '-'.
  S_IFSOCK  : 'S', # traditionally '-'.
}


_type_descs = {
  S_IFBLK   : 'block device',
  S_IFCHR   : 'character device',
  S_IFDIR   : 'directory',
  S_IFIFO   : 'named pipe',
  S_IFLNK   : 'link',
  S_IFREG   : 'file',
  S_IFSOCK  : 'socket',
}


_perm_chars = (
  ('u:', (
    (S_IRUSR, 'r'),
    (S_IWUSR, 'w'),
    (S_IXUSR, 'x'))),
  ('g:', (
    (S_IRGRP, 'r'),
    (S_IWGRP, 'w'),
    (S_IXGRP, 'x'))),
  ('o:', (
    (S_IROTH, 'r'),
    (S_IWOTH, 'w'),
    (S_IXOTH, 'x'))),
)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import re
from collections.abc import Mapping
from functools import singledispatch
from inspect import Parameter, signature
from typing import (Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Type, TypeVar, Union,
  get_args as get_type_args, get_origin, get_type_hints)

from .lex import Lexer, LexMode, LexTrans
from .parse import Choice, OneOrMore, Opt, ParseError, Parser, Struct, ZeroOrMore
from .util import memoize
from tolkien import Source, Token


_T = TypeVar('_T')


class EonContainer:
  token:Token

EonSyntax = Union[Token,EonContainer]


class EonEmpty(EonContainer):
  def __init__(self, token:Token) -> None:
    self.token = token

  def __repr__(self) -> str: return f'{type(self).__name__}({self.token})'


class EonDict(EonContainer):
  def __init__(self, token:Token, items:List[Tuple[Token,'EonSyntax']]) -> None:
    self.token = token
    self.items = items

  def __repr__(self) -> str: return f'{type(self).__name__}({self.token}, <{len(self.items)} items>)'

  def __iter__(self) -> Iterator[Tuple[Token,EonSyntax]]: return iter(self.items)


class EonList(EonContainer):
  def __init__(self, token:Token, els:List['EonSyntax']) -> None:
    self.token = token
    self.els = els

  def __repr__(self) -> str: return f'{type(self).__name__}({self.token}, <{len(self.els)} els>)'

  def __iter__(self) -> Iterator[EonSyntax]: return iter(self.els)


class EonStr(EonContainer):
  def __init__(self, token:Token, tokens:List[Token]) -> None:
    self.token = token
    self.tokens = tokens

  def __repr__(self) -> str: return f'{type(self).__name__}({self.token}, <{len(self.tokens)} tokens>)'

  def __iter__(self) -> Iterator[EonSyntax]: return iter(self.tokens)

  def value(self, source:Source) -> str:
    return ''.join(source[t] for t in self.tokens)


def _str_token_val(source:Source, token:Token) -> str:
  k = token.kind
  t = source[token]
  if k == 'esc_char': return _esc_char_vals[t]
  else: return t


_esc_char_vals = {
  '\\n' : '\n',
  '\\\\' : '\\',
  '\\"' : '"',
  "\\'" : "'",
}


class ConversionError(ParseError):
  error_prefix = 'conversion'


def parse_eon(path:str, text:str, to:Type[_T]) -> Any:
  '''
  Parse source text as EON data format.
  '''
  source = Source(name=path, text=text)
  syntax = eon_parser.parse('items', source)
  return convert_eon(syntax, source, to)


def parse_eon_or_fail(path:str, text:str, to:Type[_T]) -> Any:
  try: return parse_eon(path, text, to)
  except ParseError as e: e.fail()


@singledispatch
def convert_eon(syntax:EonSyntax, source:Source, to:Type[_T]) -> Any: # TODO: this should return _T but mypy doesn't understand.
  'Convert eon AST to a generic data value.'
  raise NotImplementedError(f'unimplemented syntax type: {syntax}')


@convert_eon.register
def convert_eon_token(syntax:Token, source:Source, to:Type[_T]) -> _T:
  text = source[syntax]
  if to is str: return text # type: ignore # TODO: handle quoted strings.
  if to is bool: return _bool_vals[text] # type: ignore
  if to in (Any, object):
    if syntax.kind == 'sym': return text # type: ignore
    if syntax.kind == 'int': return int(text) # type: ignore
    if syntax.kind == 'flt': return float(text) # type: ignore
    raise NotImplementedError(syntax)
  rtt = get_origin(to) or to
  try: return rtt(text) # type: ignore
  except Exception as e:
    raise ConversionError(source, syntax, f'expected {_fmt_type(to)}; received {syntax.kind}.\n{_fmt_exc(e)}') from e


@convert_eon.register
def convert_eon_str(syntax:EonStr, source:Source, to:Type[_T]) -> _T:
  if to not in (str, Any, object):
    raise ConversionError(source, syntax, f'expected {to}; received {syntax.token.kind}.')
  return syntax.value(source) # type: ignore


_bools = [('false', False), ('no', False), ('true', True), ('yes', True)]

_bool_vals:Dict[str,bool] = { kx : v
  for k, v in _bools
    for kx in [k, k.capitalize(), k.upper(), k[0], k[0].upper()] }


@convert_eon.register
def convert_eon_empty(syntax:EonEmpty, source:Source, to:Type[_T]) -> _T:
  if to in (Any, object, None): return None # type: ignore
  rtt = get_origin(to) or to
  try: return rtt()
  except Exception as e:
    raise ConversionError(source, syntax, f'expected {_fmt_type(to)}; received empty tree.\n{_fmt_exc(e)}') from e


@convert_eon.register
def convert_eon_list(syntax:EonList, source:Source, to:Type[_T]) -> _T:
  rtt, el_type, field_types, as_seq = _list_type_info(to)

  if field_types is None: # Use single element type.
    try:
      args = (convert_eon(el, source, el_type) for el in syntax.els)
      if as_seq: return rtt(args) # type: ignore
      else: return rtt(*args) # type: ignore
    except ConversionError as e:
      e.add_in_note(syntax.token, f'list -> {_fmt_type(to)}')
      raise
    except Exception as e:
      raise ConversionError(source, syntax,
        f'expected {_fmt_type(to)}; received list.\nElement type: {el_type}.\n{_fmt_exc(e)}') from e
  else:
    lft = len(field_types)
    if lft < len(syntax.els): # Do not tolerate excessive arguments.
      print(field_types)
      raise ConversionError(source, syntax.els[lft], f'expected {_fmt_type(to)}; received list. Argument #{lft} is excessive.')
    # If there are more fields than arguments, try it anyway; perhaps some of them have defaults. `zip` stops when els run out.
    try:
      args = (convert_eon(el, source, ft) for el, ft in zip(syntax.els, field_types))
      if as_seq: return rtt(args) # type: ignore
      else: return rtt(*args) # type: ignore
    except ConversionError as e:
      e.add_in_note(syntax.token, f'list -> {_fmt_type(to)}')
      raise
    except Exception as e:
      raise ConversionError(source, syntax, f'expected {_fmt_type(to)} structure; received list.\n{_fmt_exc(e)}') from e


@memoize
def _list_type_info(to:Type) -> Tuple[type,Any,Optional[Tuple[Any,...]],bool]:
  'Returns (rtt, el_type, field_types, as_seq).'
  if to in (Any, object): return (list, Any, None, True)
  rtt = get_origin(to) or to
  type_args = get_type_args(to)

  # Tuple types require special handling.
  if issubclass(rtt, tuple):
    lta = len(type_args)
    if lta == 2 and type_args[1] == Ellipsis: # Treat as sequence.
      el_type = type_args[0]
      return (rtt, el_type, None, True)
    elif rtt is tuple: # Treat plain tuple as struct.
      return (rtt, None, type_args, True)
    else: # Assume this is an annotated tuple subclass.
      type_hints = get_type_hints(to)
      return (rtt, None, tuple(type_hints.values()), True)

  el_type = type_args[0] if type_args else Any
  return (rtt, el_type, None, True)


@convert_eon.register
def convert_eon_dict(syntax:EonDict, source:Source, to:Type[_T]) -> _T:

  rtt:Type[_T]
  try: is_mapping, rtt, key_type, val_type, pars = _dict_type_info(to)
  except _SignatureError as e:
    msg = e.args[0]
    raise ConversionError(source, syntax, f'expected {_fmt_type(to)}; {msg}') from e

  if is_mapping:
    rtt_mapping:Callable[[Iterable],_T] = rtt
    try: return rtt_mapping((convert_eon(sk, source, key_type), convert_eon(sv, source, val_type)) for sk, sv in syntax.items)
    except ConversionError as e:
      e.add_in_note(syntax.token, f'dict -> {_fmt_type(to)}')
      raise
    except Exception as e:
      raise ConversionError(source, syntax, f'expected {_fmt_type(to)}; received dict.\n{_fmt_exc(e)}') from e
  else:
    args:Dict[str,Any] = {}
    try:
      for sk, sv in syntax.items:
        k:str = convert_eon(sk, source, str)
        try: v_type = pars[k]
        except KeyError as e:
          raise ConversionError(source, syntax, f'expected {_fmt_type(to)} ({pars}); invalid parameter: {k!r}') from e
        v = convert_eon(sv, source, v_type)
        args[k] = v
      return rtt(**args) # type: ignore
    except ConversionError as e:
      e.add_in_note(syntax.token, f'dict -> {_fmt_type(to)}')
      raise
    except Exception as e:
      raise ConversionError(source, syntax, f'expected {_fmt_type(to)} structure; received dict.\n{_fmt_exc(e)}') from e


@memoize
def _dict_type_info(to:Type) -> Tuple[bool,type,Any,Any,Dict[str,Any]]:
  'Returns (is_mapping, rtt, key_type, val_type, parameters).'
  if to in (Any, object): return (True, dict, Any, Any, {})

  # Determine if `to` is a runtime type, and get the rtt regardless.
  rtt = get_origin(to) or to

  if issubclass(rtt, Mapping):
    type_args = get_type_args(to)
    lta = len(type_args)
    if lta == 0:
      kt, vt = (Any, Any)
    elif lta == 1:
      kt = type_args[0]
      vt = Any
    else:
      kt, vt = type_args[:2]
    return (True, rtt, kt, vt, {})

  # Assume struct-like type. Get parameter types from either __init__ or __new__.
  if rtt.__init__ != _object_init:
    constructor = rtt.__init__
  elif rtt.__new__ != _object_new:
    constructor = rtt.__new__
  else:
    raise _SignatureError(f'no typed constructor information available for {rtt}')
  sig = signature(constructor)
  pars = { p.name : _parameter_type(p) for p in list(sig.parameters.values())[1:] } # Skip the 'self' or 'cls' argument.
  return (False, rtt, None, None, pars)

_object_init = object.__init__
_object_new = object.__new__


class _SignatureError(Exception): pass


def _parameter_type(p:Parameter) -> Any:
  t = p.annotation
  if p.kind != Parameter.POSITIONAL_OR_KEYWORD:
    raise _SignatureError(f'parameter kind {t.kind} not supported: {p}')
  if t is Parameter.empty: return Any
  if isinstance(t, str): raise _SignatureError(f'string parameter types not supported: {p}')
  return t


def _fmt_exc(exc:Exception) -> str:
  msg = f'{type(exc).__name__}: {exc}'
  if not msg.endswith('.'): msg += '.'
  return msg


def _fmt_type(t:Any) -> str:
  try: return str(t.__name__)
  except AttributeError: pass
  n = str(t)
  n = _typing_re.sub('', n)
  return n

_typing_re = re.compile(r'\b\w+\.')


lexer = Lexer(flags='x',
  patterns=dict(
    newline = r'\n',
    spaces  = r'\ +',
    comment = r'//[^\n]*\n',

    # TODO: section labels.

    sym = r'[A-Za-z_][0-9A-Za-z_\-\./]*',
    flt = r'[0-9]+\.[0-9]+',
    int = r'[0-9]+',

    colon = r':',
    dash = r'-',
    tilde = r'~',
    brack_o = r'\[',
    brack_c = r'\]',
    paren_o = r'\(',
    paren_c = r'\)',
    dq = r'"',
    sq = r"'",

    esc_char=r'\\[n\\"\']',
    chars_dq=r'[^\n\\"]+',
    chars_sq=r"[^\n\\']+",
  ),
  modes=[
    LexMode('main',
      kinds=[
        'newline', 'spaces', 'comment', 'sym', 'flt', 'int', 'colon', 'dash', 'tilde',
        'brack_o', 'brack_c', 'paren_o', 'paren_c', 'dq', 'sq'],
      indents=True),
    LexMode('string_dq', kinds=['chars_sq', 'esc_char', 'dq']),
    LexMode('string_sq', kinds=['chars_sq', 'esc_char', 'sq']),
  ],
  transitions=[
    LexTrans('main', kind='dq', mode='string_dq', pop='dq', consume=True),
    LexTrans('main', kind='sq', mode='string_sq', pop='sq', consume=True),
  ],
)


def _build_eon_parser() -> Parser:
  return Parser(lexer,
    dict(
      #sections=OneOrMore('section', drop='newline', transform=transform_sections),

      #section=Struct(Atom('section_label'), 'items'),

      key=Choice('flt', 'int', 'str_dq', 'str_sq', 'sym'),

      leaf=Struct(
        Choice('flt', 'int', 'str_dq', 'str_sq', 'sym'),
        'newline'),

      str_dq=Struct('dq', ZeroOrMore(Choice('esc_char', 'chars_dq')), 'dq', transform=lambda s, t, fields: EonStr(t, fields[1])),
      str_sq=Struct('sq', ZeroOrMore(Choice('esc_char', 'chars_sq')), 'sq', transform=lambda s, t, fields: EonStr(t, fields[1])),

      # Values always consume trailing newline.
      value=Choice('leaf', 'dash_list', 'tilde_dict'),


      dash_list=Struct(
        'dash',
        Choice('newline', 'value', transform=lambda s, t, label, val: None if label == 'newline' else val),
        Opt('list_body_multiline', dflt=()),
        transform=lambda s, t, fields: EonList(token=t, els=[fields[1], *fields[2]] if fields[1] else fields[2])),

      list_body_multiline=Struct('indent', OneOrMore('value', drop='newline'), 'dedent'),


      tilde_dict=Struct(
        'tilde',
        Choice('newline', 'kv_pair', transform=lambda s, t, label, pair: () if label == 'newline' else (pair,)),
        Opt('dict_body_multiline', dflt=()),
        transform=lambda s, t, fields: EonDict(token=t, items=[*fields[1], *fields[2]])),

      dict_body_multiline=Struct('indent', OneOrMore('kv_pair', drop='newline'), 'dedent'),

      kv_pair=Struct('key', 'colon', 'value'),

      # Items are sequences of either all list items, or all dict key-value pairs.
      # The returned value is inferred to be either a list or dict.
      items=ZeroOrMore('item', drop='newline', transform=transform_items),

      item=Choice('keylike_item', 'dash_list', 'tilde_dict', transform=transform_item),

      keylike_item=Struct('key', 'newline_or_colon_body'),

      newline_or_colon_body=Choice('newline', 'colon_body',
        transform=lambda s, t, label, val: None if label == 'newline' else val), # Newline implies that the just-parsed key is a list element.

      colon_body=Struct('colon', 'body'),

      body=Choice('value', 'body_multiline'),

      body_multiline=Struct('newline', 'indent', 'items', 'dedent'),
    ),
  literals=('newline', 'indent', 'dedent', 'colon'),
  drop=('comment', 'spaces'))


# Parser transformers.


def transform_items(source:Source, start:Token, items:List[Tuple[Any,Any]]) -> EonContainer:
  is_dict:bool
  vals:List[Any] = []
  for token, p in items:
    k, v = p
    is_pair = (v is not None) # None is never returned by value transformers, so it is a sufficient discriminator.
    if not vals:
      is_dict = is_pair
    elif is_dict != is_pair: # Inconsistent.
      if is_pair:
        msg = f'Expected list elements; received key-value pair.'
      else:
        msg = f'Expected key-value pair; received list element.'
      raise ParseError(source, token, 'inconsistent sequence. ' + msg, notes=[(start, 'note: first element is here.')])
    vals.append(p if is_pair else k)
  if not vals:
    return EonEmpty(token=start)
  if is_dict:
    return EonDict(token=start, items=vals)
  else:
    return EonList(token=start, els=vals)


def transform_item(source:Source, token:Token, label:str, item:Any) -> Tuple[Token,Any]:
  if label == 'keylike_item':
    assert isinstance(item, tuple) and len(item) == 2
    return (token, item)
  else: return (token, (item, None))


eon_parser = _build_eon_parser()


def main() -> None:
  '''
  Parse specified files (or stdin) as EON and print each result.'
  '''
  from sys import argv
  from .io import outD

  args = argv[1:] or ['/dev/stdin']
  for path in args:
    with open(path) as f:
      text = f.read()
    try: obj:Any = parse_eon(path, text, to=object)
    except ParseError as e: e.fail()
    outD(path, obj)


if __name__ == '__main__': main()
